/**
 * @file condVar.cpp
 *
 * @author Jan Hadrava
 * @ingroup iopccommon
 */

#include <pthread.h>
#include <errno.h>
#include "condVar.h"
#include "trace.h"

namespace iopc {

CondVar::CondVar()
	throw (IopcNoMemoryException, IopcRsrcLackException)
{
	int res;
	pthread_mutexattr_t attr;

	switch (pthread_mutexattr_init(&attr)) {
	case 0:
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}

	switch (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK)) {
	case 0:
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}

	switch (pthread_mutex_init(&mutex, &attr)) {
	case 0:
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EAGAIN:
	case EPERM:
		res = pthread_mutexattr_destroy(&attr); IOPC_ASSERT(res == 0);
		IOPC_LOG_THROW(IopcRsrcLackException());
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}

	switch (pthread_mutexattr_destroy(&attr)) {
	case 0:
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
	switch (pthread_cond_init(&cond, NULL)) {
	case 0:
		return;
	break;
	case ENOMEM:
		res = pthread_mutex_destroy(&mutex); IOPC_ASSERT(res == 0);
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EAGAIN:
		res = pthread_mutex_destroy(&mutex); IOPC_ASSERT(res == 0);
		IOPC_LOG_THROW(IopcRsrcLackException());
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

CondVar::~CondVar()
	throw ()
{
	int res;
	res = pthread_cond_destroy(&cond); IOPC_ASSERT(res == 0);
	res = pthread_mutex_destroy(&mutex); IOPC_ASSERT(res == 0);
}

void CondVar::lock()
	throw (IopcRsrcLackException, IopcDeadlockException)
{
	switch (pthread_mutex_lock(&mutex)) {
	case 0:
		return;
	break;
	case EAGAIN:
		IOPC_LOG_THROW(IopcRsrcLackException());
	break;
	case EDEADLK:
		IOPC_LOG_THROW(IopcDeadlockException());
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

void CondVar::unlock()
	throw (IopcIncorrectUsageException)
{
	switch (pthread_mutex_unlock(&mutex)) {
	case 0:
		return;
	break;
	case EPERM:
		IOPC_LOG_THROW(IopcIncorrectUsageException());
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

void CondVar::signal()
	throw ()
{
	switch (pthread_cond_broadcast(&cond)) {
	case 0:
		return;
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

void CondVar::wait()
	throw ()
{
	switch (pthread_cond_wait(&cond, &mutex)) {
	case 0:
		return;
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

void CondVar::wait(const TimeStamp& ts)
	throw ()
{
	struct timespec tspec = ts;	/* using cast operator */
	int res;
	switch (res = pthread_cond_timedwait(&cond, &mutex, &tspec)) {
	case 0:
	case ETIMEDOUT:
	case EINVAL:
		return;
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

}
