/**
* @file trace.h
*
* @author Michal Kopecky
* @author Jan Hadrava
* @author Petr Cermak
*
* @brief @a LogWriter singleton and logging/tracing/debugging macros.
*
* IOPC_DEBUG macro enables or disables all debugging messages
* IOPC_TRACE macro enables or disables all tracing messages.
* Release versions of IOPC 2 should have these macros undefined.
*
* The logging macros can be used in the following way:
* @code
* IOPC_DEBUG macros (if IOPC_DEBUG is undefined, no messages are logged):
* IOPC_DEBUGLOG(msg)
*
* IOPC_TRACE macros (if IOPC_TRACE is undefined, no messages are logged):
* void func() {
*	IOPC_FUNC_INVOKED;
*	 ..
*	IOPC_FUNC_ENDED;
*	or
*	IOPC_FUNC_END(string retval)
* }
*
* IOPC_LOG_THROW(IopcDbException("error"))
*
* IOPC_LOG_CATCH(IopcException& exc)
*	IOPC_LOG_THROW(exc);
* IOPC_LOG_CATCH_NEXT(...)
*	IOPC_UNREACHABLE;
* IOPC_LOG_CAUGHT
*
* Always logging macros:
* IOPC_ASSERT(cond); // Exits program if condition evaluates to false
* IOPC_UNREACHABLE;  // Exists program if reached
* @endcode
*
* @ingroup iopccommon
*/

#ifndef IOPCTRACE_H
#define IOPCTRACE_H

#include <fstream>
#include <ostream>
#include <string>
#include "defs.h"
#include "mutex.h"
#include <cstdlib>

// ----- Comment or uncomment logging options
#ifdef ECLIPSE_DEBUG
// Debug messages are enabled
#define IOPC_DEBUG
// Trace and SQL trace messages are enabled
#define IOPC_TRACE
#endif

namespace iopc
{

/**
 * @brief A @a LogWriter singleton is used to log messages for the IOPC 2 library or user application
 *
 * @a LogWriter writes the logged messages to standard output and to a log file if file logging is enabled.
 *
 * Logger logs messages of various severity levels. The levels are defined by the @a MsgType enumeration.
 * Message can be filtered by these levels (message types) and their combinations.
 *
 * Log macros in the @a iopccommon/trace.h file use @a LogWriter to write log messages.
 *
 * Logging methods are thread safe.
 *
 * @ingroup iopccommon
 */
class DLL_EXTERNAL LogWriter
{
public:
	/**
	 * @brief Log message severity level (message type)
	 */
	enum MsgType {
		NONE = 0, 		/**< Turns logging off */
		/**
		 * @brief SQL trace messages representing SQL statements that has been sent to the underlying database.
		 *
		 * Messages in this category are exclusively generated by the @c iopcdb library.
		 */
		INFO_SQL = 1,
		/**
		 *  @brief Informational message.
		 *
		 *  Can be generated by a user application.
		 */
		INFO = 2,
		/**
		 * @brief Warning message
		 *
		 * Can be generated by a user application.
		 */
		WARNING = 4,
		/**
		 * @brief Messages generated usually as a result of an exception that prevented required operation to be finished.
		 *
		 * Can be generated by the user application.
		 */
		ERROR = 8,
		/**
		 * @brief Messages that trace the library execution.
		 *
		 * Turning this filter on may result in large amount of logged messages.
		 * May be also disabled by commenting out the IOPC_TRACE macro in the @a iopccommon/trace.h file.
		 * Release versions of IOPC 2 should have this macro undefined.
		 */
		TRACE = 16,
		/**
		 * @brief Messages useful for debugging purposes.
		 *
		 * May be also disabled by commenting out the IOPC_DEBUG macro in the @a iopccommon/trace.h file.
		 * Release versions of IOPC 2 should have this macro undefined.
		 */
		DEBUG = 32};

	/**
	 * @brief Default log file name - iopc_log.txt
	 */
	static const std::string defaultFileName;
	/**
	 * @brief Default file log message mask.
	 *
	 * Defined as:
	 * @code
	 * LogWriter::INFO | LogWriter::WARNING | LogWriter::ERROR
	 * IOPC_TRACE defined: LogWriter::INFO_SQL | LogWriter::INFO | LogWriter::WARNING | LogWriter::ERROR | LogWriter::TRACE
	 * IOPC_DEBUG defined: LogWriter::INFO | LogWriter::WARNING | LogWriter::ERROR | LogWriter::DEBUG
	 * IOPC_DEBUG and IOPC_TRACE defined: LogWriter::INFO_SQL | LogWriter::INFO | LogWriter::WARNING | LogWriter::ERROR | LogWriter::DEBUG | LogWriter::TRACE
	 * @endcode
	 */
	static const int defaultTraceMask;

	/**
	 * @brief Default console log message mask.
	 *
	 * Defined as:
	 * @code
	 * LogWriter::INFO | LogWriter::WARNING | LogWriter::ERROR
	 * IOPC_TRACE defined: LogWriter::INFO | LogWriter::WARNING | LogWriter::ERROR
	 * IOPC_DEBUG defined: LogWriter::INFO | LogWriter::WARNING | LogWriter::ERROR | LogWriter::DEBUG
	 * IOPC_DEBUG and IOPC_TRACE defined: LogWriter::INFO | LogWriter::WARNING | LogWriter::ERROR | LogWriter::DEBUG
	 * @endcode
	 */
	static const int defaultTraceMaskTTY;

	/**
	 * @brief Log file log message mask
	 */
	int traceMask;
	/**
	 * @brief Console log message mask
	 */
	int traceMaskTTY;

	LogWriter();
	virtual ~LogWriter();

	void logMsgInst(MsgType mask, const std::string& message);
	void logLineInst(MsgType mask, std::string file, unsigned int line, std::string function, const std::string& message);

	void logOpenInst(const std::string& fileName = defaultFileName, bool append = true);
	void logCloseInst();
	void logFlushInst();

	/**
	 * @brief Returns a @a MsgType string representation
	 */
	std::string getMsgTypeStr(MsgType msgType);

	/**
	 * @brief Initializes the @a LogWriter
	 *
	 * Called by @a IopcCommon::init()
	 */
	static void initInstance();

	/**
	 * @brief Terminates the @a LogWriter
	 *
	 * Called by @a IopcCommon::shutdown()
	 */
	static void terminateInstance();

	/**
	 * @brief Returns the singletion instance
	 */
	static LogWriter& getInstance();

	/**
	 * @brief Logs a message
	 *
	 * @param mask Message type/level
	 * @param message The message
	 */
	static inline void logMsg(MsgType mask, const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsgInst(mask, message);
	};

	/**
	 * @brief Logs a trace message. Used by tracing macros.
	 */
	static inline void logLine(MsgType mask, std::string file, unsigned int line, std::string function, const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logLineInst(mask, file, line, function, message);
	};

	/**
	 * @brief Opens the log file.
	 *
	 * Called by @a IopcCommon::init()
	 */
	static inline void logOpen(const std::string& fileName = defaultFileName, bool append = true) {
		if (!isInitialized()) return;
		getInstance().logOpenInst(fileName, append);
	};

	/**
	 * @brief Closes the log file.
	 *
	 * Called by @a IopcCommon::shutdown()
	 */
	static inline void logClose() {
		if (!isInitialized()) return;
		getInstance().logFlushInst(); getInstance().logCloseInst();
	};

	/**
	 * @brief Logs a @a LogWriter::INFO message
	 *
	 * @param message The message
	 */
	static inline void logInfo(const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsg(INFO, message);
	};

	/**
	 * @brief Logs a @a LogWriter::WARNING message
	 *
	 * @param message The message
	 */
	static inline void logWarning(const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsg(WARNING, message);
	};

	/**
	 * @brief Logs a @a LogWriter::ERROR message
	 *
	 * @param message The message
	 */
	static inline void logError(const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsg(ERROR, message);
	};

	/**
	 * @brief Logs a @a LogWriter::INFO_SQL message
	 *
	 * @param message The message
	 */
	static inline void logSql(const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsg(INFO_SQL, message);
	};

	/**
	 * @brief Returns true if the @a LogWriter singletion is initialized
	 */
	static inline bool isInitialized() { return instance != NULL; }
private:
	static Mutex logLock;
	static LogWriter* instance;
	std::string fileName;

	std::ofstream logStream;
	//int va_log( MsgType mask, const char *format, va_list argp );
};

//extern LogWriter IopcLogWriter;


#ifndef __FUNCTION__
#ifdef __func__
#define __FUNCTION__ __func__
#endif
#else
#define __FUNCTION__ ""
#endif


// ----- DEBUG
#ifndef IOPC_DEBUG
#define IOPC_DEBUGLOG(msg) ;
#else // #ifndef IOPC_DEBUG
#define IOPC_DEBUGLOG(msg) LogWriter::logLine(LogWriter::DEBUG, __FILE__, __LINE__, __FUNCTION__, msg);
#endif // #ifndef IOPC_DEBUG

// ----- TRACE
#ifdef IOPC_TRACE
#define IOPC_FUNC_INVOKED						\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, "Function invoked");
#define IOPC_FUNC_END(retval)								\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Function ends with return value: " + #retval)
#define IOPC_FUNC_ENDED								\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Function ended")
#define IOPC_LOG_THROW(exc)								\
	do {									\
	const std::exception& e = exc;	\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Throwing exception: " + #exc + " what: " + e.what());					\
	throw exc;							\
	} while (0)
#define IOPC_LOG_RETHROW(exc)	\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Rethrowing exception: " + #exc);	\
	throw ;
#define IOPC_LOG_CATCH(exc)							\
	catch (exc) {							\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Exception caught: " + #exc);
#define IOPC_LOG_CAUGHT							\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, "Exception solved");		\
	}
#else // #ifdef IOPC_TRACE
#define IOPC_LOG_THROW(exc)								\
	do {									\
	throw exc;							\
	} while (0)
#define IOPC_FUNC_INVOKED ;
#define IOPC_FUNC_END(retval) ;
#define IOPC_FUNC_ENDED ;
#define IOPC_LOG_RETHROW(exc)	\
	throw;
#define IOPC_LOG_CATCH(exc)	\
catch (exc) {
#define IOPC_LOG_CAUGHT	\
	}
#endif	// #ifdef IOPC_TRACE



#define IOPC_LOG_CATCH_NEXT(exc)						\
	IOPC_LOG_CAUGHT							\
	IOPC_LOG_CATCH(exc)


// ----- Common macros
#define IOPC_ASSERT(cond)									\
	do {											\
	if (!(cond)) {										\
	LogWriter::logLine(LogWriter::ERROR, __FILE__, __LINE__, __FUNCTION__, (std::string) "ASSERT failed: " + #cond);				\
	exit(-1);									\
	}											\
	} while (0)
#define IOPC_UNREACHABLE									\
	do {											\
	LogWriter::logLine(LogWriter::ERROR, __FILE__, __LINE__, __FUNCTION__, "Unreachable code reached");				\
	exit(-1);									\
	} while (0)



} /*namespace polite*/

#endif //__L_TRACE_H__
