/**
 * @file util.h
 *
 * @author Petr Cermak
 *
 * @brief Various, mostly string manipulation utilities
 *
 * @ingroup iopccommon
 */

#ifndef UTIL_H_
#define UTIL_H_

#include <map>
#include <string>
#include <vector>
#include "defs.h"

namespace iopc {

#define IOPC_PARAM_DELIMITER ";"

/**
 * @brief Various string manipulation utilities
 * @ingroup iopccommon
 */
class DLL_EXTERNAL StringUtil {

public:
	/**
	 * @brief Splits a @a delimiter delimited string into a vector of strings
	 *
	 * @param input The string containing @a delimiter delimited values
	 * @param delimiter The delimiting string
	 * @param results [out] Vector containing the substrings from @a input
	 * @param skipEmpty Determines whether to skip empty substrings (two delimiters in sequence). Default is false.
	 * @author Paul J. Weiss
	 */
	static int splitString(const std::string& input, const std::string& delimiter, std::vector<std::string>& results, bool skipEmpty = false);

	/**
	 * @brief Concatenates strings using a @a delimiter
	 *
	 * @param strings The list of strings to be concatenated
	 * @param delimiter The strings are concatenated using this delimiter.
	 *
	 * @returns Concatenated string.
	 */
	static std::string joinStrings(std::vector<std::string> strings, const std::string& delimiter);

	/**
	 * @brief Converts a string of format "key1=value1;key2=value2" into a STL map
	 *
	 * Key-value pair delimiter is specified by the IOPC_PARAM_DELIMITER macro defined
	 * in @a iopccommon/util.h
	 *
	 * @param paramStr The string to be converted
	 * @returns The map containing keys and values from the input string
	 */
	static std::map<std::string, std::string> createParamMap(const std::string& paramStr);

	/**
	 * @brief Concatenates two strings (no delimiter)
	 */
	static std::string join(const char* string1, const char* string2);

	/**
	 * @brief Removes file extension from a file name
	 *
	 * @param fileName The file name. May include a path
	 * @returns The file name without extension.
	 */
	static inline std::string removeFileExt(std::string fileName) {
		return fileName.substr(0, fileName.find_last_of('.'));
	}

	/**
	 * @brief Parses integer from a string
	 *
	 * @return Parsed integer. If the string couldn't be parsed, the behaviour is undefined.
	 */
	static int parseInt(const std::string& input);

	/**
	 * @brief Parses a boolean value from a string
	 *
	 * @return true if the input value is "true", false otherwise
	 */
	static bool parseBool(const std::string& input);

	/**
	 * @brief Converts an integer value to string
	 */
	static std::string toString(int i);

};

/**
 * @brief Various utilities
 *
 * @ingroup iopccommon
 */
class DLL_EXTERNAL Util {
public:
	/**
	 * @brief Returns a string representing current date and time.
	 *
	 * Format of the string is YYMMDD HH24:mm:SS
	 */
	static std::string currentTime();
};

}

#endif /*UTIL_H_*/
