/**
 * @file driver.cpp
 * 
 * @author Petr Cermak
 */

#include "driver.h"
#include "driverManager.h"
#include "basicDatabase.h"
#include "driverExtension.h"

namespace iopc {

Driver::Driver(const std::string& driverName)
{
	this->name = driverName;
	DriverManager::getInstance().registerDriver(this, driverName);
}


Driver::~Driver()
{
	
}

void Driver::init(std::map<std::string, DriverExtension*>& extensions)
{
	for (std::map<std::string, DriverExtension*>::const_iterator it = extensions.begin(); it != extensions.end(); it++) {
		registerExtension(it->first, *it->second);
	}
	doInit();
	for (std::map<std::string, DriverExtension*>::const_iterator it = extensions.begin(); it != extensions.end(); it++) {
		it->second->doInit(*this);
	}
}

void Driver::shutdown()
{
	if (databases.size() > 0) {
		LogWriter::logWarning("Removing registered databases from driver " + name);
		disconnectDatabases();
	}
	doShutdown();
}

Database* Driver::getDatabase(const std::string& dbName)
{
	Database* db = new BasicDatabase(getDatabaseImpl(dbName), *this);
	mutex.lock();
	addDatabase(db);
	mutex.unlock();
	return db;
}

void Driver::returnDatabase(Database* db)
{
	db->disconnect();
	returnDatabaseImpl(db->getImplementation());
	mutex.lock();
	removeDatabase(db->getBasicDatabase());
	mutex.unlock();
	delete db;
}

void Driver::addDatabase(Database* db)
{
	databases.insert(db);
}

void Driver::removeDatabase(Database* db)
{
	DatabaseSetIterator iter;
	
	iter = databases.find(db);
	if (iter != databases.end())
		databases.erase(iter);
}

void Driver::disconnectDatabases()
{
	for (DatabaseSetIterator iter = databases.begin(); iter != databases.end(); ++iter) {
		(*iter)->disconnect();
		returnDatabaseImpl((*iter)->getImplementation());
		delete (*iter);
	}
	databases.clear();
}


}
