/**
 * @file oracle10gCursorImpl.cpp
 *
 * @author Josef Troch Josef Troch (<a href="mailto:josef.troch@email.cz">josef.troch@email.cz</a>) 2003-2004
 * @author Petr Cermak
 */


#include "oracle10gCursorImpl.h"
#include "iopccommon/trace.h"
#include "iopccommon/exceptionBase.h"
#include <string>
#include <utility>
#include "iopcmeta/typeDesc.h"

using namespace std;

namespace iopc {

//! Oracle error code: %Database Lock - presicely: ORA-00054: resource busy and acquire with NOWAIT specified.
// TODO: kouknout jeslti neni nekde ve .h souborech od OCI
#define ORA_RESOURCE_BUSY 54


Oracle10gCursorImpl::Oracle10gCursorImpl(OCIEnv* ociEnv, OCIError* ociErr, OCISvcCtx* ociSvc, bool isAutoCommit, const std::string& sqlCommand)
{
	IOPC_FUNC_INVOKED;

	ociEnvironment = ociEnv;
	errorHandle = ociErr;
	serviceContext = ociSvc;
	autoCommit = isAutoCommit;
	this->sqlCommand = sqlCommand;
	IOPC_FUNC_ENDED;
}

Oracle10gCursorImpl::~Oracle10gCursorImpl()
{
	IOPC_FUNC_INVOKED;
	clear();
	IOPC_FUNC_ENDED;
}


void Oracle10gCursorImpl::clear()
{
	IOPC_FUNC_INVOKED;
	map<int, IndicatorStruct*>::iterator it;
	IOPC_DEBUGLOG("1");
	for (it = indicatorVariables.begin(); it != indicatorVariables.end(); it++) {
		IOPC_DEBUGLOG("1a");
		delete it->second->indicator;
		IOPC_DEBUGLOG("1b");
		delete it->second;
		IOPC_DEBUGLOG("1c");
	}
	IOPC_DEBUGLOG("2");
	indicatorVariables.clear();
	IOPC_FUNC_ENDED;
}

void Oracle10gCursorImpl::open()
{
	IOPC_FUNC_INVOKED;
	position = 0;

	sword ret;
	//create statement handle
	ret = OCIHandleAlloc(ociEnvironment, (dvoid **)&stmtHandle, OCI_HTYPE_STMT, 0, NULL);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIHandleAlloc");
}

void Oracle10gCursorImpl::prepare()
{
	IOPC_FUNC_INVOKED;
	IOPC_DEBUGLOG(string("Sql: ") + this->sqlCommand);

	sword ret;

	//prepare the statement
	// TODO konfigurace mode.
	ret = OCIStmtPrepare(stmtHandle, errorHandle, (OraText *)sqlCommand.c_str(), sqlCommand.length(), OCI_NTV_SYNTAX, OCI_DEFAULT);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIStmtPrepare");

	//is it a SELECT or some other SQL statement?
	ub2 stmtType;
	ret = OCIAttrGet(stmtHandle, OCI_HTYPE_STMT, &stmtType, NULL, OCI_ATTR_STMT_TYPE, errorHandle);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIAttrGet");

	isSelectStatement = (stmtType==OCI_STMT_SELECT);

	IOPC_FUNC_ENDED;
}

void Oracle10gCursorImpl::preExecBind(const int pos, const void* varAddr, unsigned int size, const Type& varType, bool setNull,  const MetadataHolder& attributes)
{
	IOPC_FUNC_INVOKED;
	static sb2 nullIndicator = -1; //null value
	static sb2 okIndicator = 0;		//common value
	sword ret;

	//int len = varType.getSizeOf(varAddr);
	ub2 type; //Oracle external type of variable

	if (varAddr == NULL) setNull = true;

	type = (ub2) varType.getMeta("db.oracle.type.otype").getIntValue();  //getOracleType(varType);
	/*if (type == UNKNOWN_DATA_TYPE) {
		if (!setNull) {
			IOPC_LOG_THROW(IopcDbException("Oracle 10g Driver can handle only IOPC_TYPECLASS_SIMPLE or IOPC_TYPECLASS_STRING datatypes."));
		}
		else type = SQLT_INT; //Oracle needs "valid" column type also if setting a value to NULL...
	}*/

	//bind the variable
	sb2 * indicatorPtr = (setNull) ? &nullIndicator : &okIndicator;
	OCIBind* bindHandle = NULL;
	ret = OCIBindByPos(stmtHandle, &bindHandle, errorHandle, pos, (dvoid *)varAddr, size,
											 type, indicatorPtr, NULL, NULL, 0, 0, OCI_DEFAULT);

	IOPC_HANDLE_OCI_ERROR(ret, "OCIBindByPos");

	if (!setNull && varType == TypeDesc<std::wstring>::getType()) //tell OCI to use Unicode
	{
		ub2 csid = OCI_UCS2ID;
		ret = OCIAttrSet(bindHandle, OCI_HTYPE_BIND, &csid, 0, OCI_ATTR_CHARSET_ID, errorHandle);
		IOPC_HANDLE_OCI_ERROR(ret, "OCIAttrSet");

		sb2 wcharCount = (short)wcslen((wchar_t*)varAddr);
		ret = OCIAttrSet(bindHandle, OCI_HTYPE_BIND, &wcharCount, 0, OCI_ATTR_MAXCHAR_SIZE, errorHandle);
		IOPC_HANDLE_OCI_ERROR(ret, "OCIAttrSet");
	}
}


void Oracle10gCursorImpl::execute()
{
	IOPC_FUNC_INVOKED;

	sword ret;

	ub4 mode = (autoCommit) ? OCI_COMMIT_ON_SUCCESS : OCI_DEFAULT;

	//execute the statement
	ret = OCIStmtExecute(serviceContext, stmtHandle, errorHandle, (isSelectStatement) ? 0 : 1, 0, NULL, NULL, mode);
	if (!OCI_OK(ret)) {
		pair<string, sb4> err = getOCIError(ret, "OCIStmtExecute");
		if (err.second == ORA_RESOURCE_BUSY) {
			IOPC_LOG_THROW(IopcDbDeadlockException("DB Deadlock. ORA_RESOURCE_BUSY returned."));
		} else {
			IOPC_LOG_THROW(IopcDbException(err.first + " Sql: " + this->sqlCommand));
		}
	} else if (OCI_MSG(ret)) {
		LogWriter::logInfo(getOCIMsg(ret, getOCIMsg(ret, "OCIStmtExecute")));
	}

	IOPC_FUNC_ENDED;
}

void Oracle10gCursorImpl::preFetchBind(const int pos, const void* varAddr, unsigned int size, const Type& varType, bool* isNull, const MetadataHolder& attributes)
{
	sword ret;
	ub2 type; //Oracle external type of variable
	//int len = varType.getSizeOf(varAddr);

	type = (ub2) varType.getMeta("db.oracle.type.otype").getIntValue(); //type = getOracleType(varType);
	/*if (type == UNKNOWN_DATA_TYPE) {
		IOPC_LOG_THROW(IopcDbException("Oracle 10g Driver can handle only IOPC_TYPECLASS_SIMPLE or IOPC_TYPECLASS_STRING datatypes."));
	}*/

	//create indicator variable struct
	map<int, IndicatorStruct*>::const_iterator it = indicatorVariables.find(pos);
	IndicatorStruct* ind;
	if (it == indicatorVariables.end()) {
		ind = new IndicatorStruct;
		ind->ociDef = NULL;
		ind->indicator = new sb2;
		indicatorVariables[pos] = ind;
	} else {
		ind = it->second;
	}
	ind->boolPtr = isNull;

	//OCIDefine* defineHandle = NULL;
	//defineHandle allocated by the following statement is implicitly freed when destroying the stmtHandle.
	ret = OCIDefineByPos(stmtHandle, &(ind->ociDef), errorHandle, pos, (dvoid *)varAddr, size, type, ind->indicator,
											 NULL, NULL, OCI_DEFAULT);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIDefineByPos");

	if (varType == TypeDesc<std::wstring>::getType()) {
		ub2 csid = OCI_UCS2ID;
		ret = OCIAttrSet(ind->ociDef, OCI_HTYPE_DEFINE, &csid, 0, OCI_ATTR_CHARSET_ID, errorHandle);
		IOPC_HANDLE_OCI_ERROR(ret, "OCIAttrSet");
	}

	IOPC_FUNC_ENDED;
}

bool Oracle10gCursorImpl::fetchNext()
{
	IOPC_FUNC_INVOKED;

	sword ret;

	//fetch next row
	ret = OCIStmtFetch(stmtHandle, errorHandle, 1, OCI_FETCH_NEXT, OCI_DEFAULT);
	if (ret==OCI_NO_DATA) {
			return false;
	}
	IOPC_HANDLE_OCI_ERROR(ret, "OCIStmtFetch");

	//process indicator variables
	map<int, IndicatorStruct*>::iterator it;
	for (it = indicatorVariables.begin(); it != indicatorVariables.end(); it++) {
		if (it->second->boolPtr != NULL) {
			*(it->second->boolPtr) = (*(it->second->indicator) == -1) ? true : false;
		}
	}

	position++;
	return true;
}


bool Oracle10gCursorImpl::fetchPrev()
{
	IOPC_LOG_THROW(IopcNotSupportedException("doFetchPrev is not supported"));
}


long Oracle10gCursorImpl::getPosition()
{
	IOPC_FUNC_INVOKED;
	return position;
}

void Oracle10gCursorImpl::close()
{
	IOPC_FUNC_INVOKED;

	sword ret = OCIHandleFree(stmtHandle, OCI_HTYPE_STMT);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIHandleFree");

	clear();

	IOPC_FUNC_ENDED;
}


/*
void Oracle10gCursorImpl::test()
{
	//OCIStmt* stmtHandle2;
	sword ret;
	int* var = new int;
	char*  select = "select 3 from dual";

	LogWriter::logError("start");

	//create statement handle
	ret = OCIHandleAlloc(ociEnvironment, (dvoid **)&stmtHandle, OCI_HTYPE_STMT, 0, NULL);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIHandleAlloc");

	LogWriter::logError("1");
	//prepare the statement
	ret = OCIStmtPrepare(stmtHandle, errorHandle, (OraText *) select, 18, OCI_NTV_SYNTAX, OCI_DEFAULT);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIStmtPrepare");

	LogWriter::logError("2");
	ret = OCIStmtExecute(serviceContext, stmtHandle, errorHandle, 0, 0, NULL, NULL, OCI_DEFAULT);
	if (!OCI_OK(ret)) {
		pair<string, sb4> err = getOCIError(ret, "OCIStmtExecute");
		if (err.second == ORA_RESOURCE_BUSY) {
			// todo udelat deadlock zdedeny z dbE.
			IOPC_LOG_THROW(IopcDeadlockException("DB Deadlock. ORA_RESOURCE_BUSY returned."));
		} else {
			IOPC_LOG_THROW(IopcDbException(err.first));
		}
	} else if (OCI_MSG(ret)) {
		LogWriter::logInfo(getOCIMsg(ret, getOCIMsg(ret, "OCIStmtExecute")));
	}

	LogWriter::logError("3");
	OCIDefine* defineHandle = NULL;
	//defineHandle allocated by the following statement is implicitly freed when destroying the stmtHandle.
	ret = OCIDefineByPos(stmtHandle, &defineHandle, errorHandle, 1, (dvoid *)var, sizeof(int), SQLT_INT, NULL,
											 NULL, NULL, OCI_DEFAULT);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIDefineByPos");

	LogWriter::logError("4");
	ret = OCIStmtFetch2(stmtHandle, errorHandle, 1, OCI_FETCH_NEXT, 1, OCI_DEFAULT);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIStmtFetch2");

	LogWriter::logError(string("5") + char('0' + var));
	ret = OCIHandleFree(stmtHandle, OCI_HTYPE_STMT);
	IOPC_HANDLE_OCI_ERROR(ret, "OCIHandleFree");
}
*/

}
