#include "oracle10gObjectStatementsFeature.h"
#include <string>
#include "iopcmeta/type.h"
#include "iopccommon/globalSettings.h"
#include <sstream>
#include "iopclib/types/dbType.h"
#include <set>
#include "iopccommon/util.h"
//#include <list>

using namespace iopc;
using namespace std;

bool Oracle10gObjectStatementsFeature::validate(ObjectMappingStrategy& currentStrategy, string& msg)
{
	IOPC_DEBUGLOG(currentStrategy.getDbType().getType().getQualifiedName());
	int persistentParents = 0;
	for (Type::ParentsIterator it = currentStrategy.getDbType().getType().getParents().begin(); it != currentStrategy.getDbType().getType().getParents().end(); it++) {
		if ((*it)->isTypeOf<DatabaseObject>()) {
			persistentParents++;
		}
	}
	if (persistentParents > 1) {
		msg = "Oracle object mapping cannot handle multiple inheritance. Type: " + currentStrategy.getDbType().getType().getQualifiedName();
		return false;
	}
	return true;
}

string Oracle10gObjectStatementsFeature::getInsertStatement(ObjectMappingStrategy& currentStrategy,  const MetadataHolder& attributes, bool useCache)
{
	if (useCache && cache.hasCachedStatement(currentStrategy.getType().getQualifiedName() + cache.insertStatementType)) {
		return cache.getCachedStatement(currentStrategy.getType().getQualifiedName() + cache.insertStatementType);
	}

	vector<const MetadataHolder*> cols;
	cols.push_back(&currentStrategy.getOidColumn());
	for (ObjectMappingStrategy::PersistentAttributesIterator it = currentStrategy.getMappedAttributes().begin(); it != currentStrategy.getMappedAttributes().end(); it++) {
		//if (**it == currentStrategy.getOidColumn()) continue;
		cols.push_back(*it);
	}

	ostringstream ostr;

	ostr << "BEGIN " << currentStrategy.getType().getMeta("db.type").getStringValue() << ".insert_object(";
	ostr << sqlStatements->getColumnsBindStr(cols, true, "p_") << "); END;";

	if (useCache) {
		cache.setCachedStatement(currentStrategy.getType().getQualifiedName() + cache.insertStatementType, ostr.str());
	}
	return ostr.str();
}

string Oracle10gObjectStatementsFeature::getUpdateStatement(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes, bool useCache)
{
	if (useCache && cache.hasCachedStatement(currentStrategy.getType().getQualifiedName() + cache.updateStatementType)) {
		return cache.getCachedStatement(currentStrategy.getType().getQualifiedName() + cache.updateStatementType);
	}

	vector<const MetadataHolder*> cols;
	for (ObjectMappingStrategy::PersistentAttributesIterator it = currentStrategy.getMappedAttributes().begin(); it != currentStrategy.getMappedAttributes().end(); it++) {
		cols.push_back(*it);
	}
	cols.push_back(&currentStrategy.getOidColumn()); // put the oid column to the end to simulate the where binding position

	ostringstream ostr;

	ostr << "BEGIN " << currentStrategy.getType().getMeta("db.type").getStringValue() << ".update_object(";
	ostr << sqlStatements->getColumnsBindStr(cols, true, "p_") << "); END;";

	if (useCache) {
		cache.setCachedStatement(currentStrategy.getType().getQualifiedName() + cache.updateStatementType, ostr.str());
	}
	return ostr.str();
}

string Oracle10gObjectStatementsFeature::getDeleteStatement(ObjectMappingStrategy& currentStrategy,  const MetadataHolder& attributes, bool useCache)
{
	if (useCache && cache.hasCachedStatement(currentStrategy.getType().getQualifiedName() + cache.deleteStatementType)) {
		return cache.getCachedStatement(currentStrategy.getType().getQualifiedName() + cache.deleteStatementType);
	}
	ostringstream ostr;
	ostr << "BEGIN " << currentStrategy.getType().getMeta("db.type").getStringValue() << ".delete_object(:oid); END;";

	if (useCache) {
		cache.setCachedStatement(currentStrategy.getType().getQualifiedName() + cache.deleteStatementType, ostr.str());
	}
	return ostr.str();
}


string Oracle10gObjectStatementsFeature::getSelectFromClassViewStatement(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes, bool useCache)
{
	if (useCache && cache.hasCachedStatement(currentStrategy.getType().getQualifiedName() + cache.selectViewStatementType)) {
		return cache.getCachedStatement(currentStrategy.getType().getQualifiedName() + cache.selectViewStatementType);
	}

	MetadataHolder tableMeta;
	MetadataHolder selectMeta;
	vector<const MetadataHolder*> cols;
	vector<const MetadataHolder*> tables;
	string whereStr;
	string ret;

	tableMeta["db.table"].setStringValue(currentStrategy.getType().getMeta("db.view.sv").getStringValue());
	tables.push_back(&tableMeta);

	for (DbType::PersistentAttributesIterator it = currentStrategy.getDbType().getAllPersistentAttributes().begin(); it != currentStrategy.getDbType().getAllPersistentAttributes().end(); it++) {
		MetadataHolder* colMeta = new MetadataHolder();
		(*colMeta)["db.column"].setStringValue((*it)->getMeta("db.viewColumn").getStringValue());
		cols.push_back(colMeta);
	}

	selectMeta["db.select.where"].setStringValue(getWhere(currentStrategy.getDbType().getKeyAttributes(), "db.viewColumn"));

	ret = sqlStatements->getSelectStatement(tables, cols, selectMeta);

	for (vector<const MetadataHolder*>::const_iterator it = cols.begin(); it != cols.end(); it++) {
		delete *it;
	}

	if (useCache) {
		cache.setCachedStatement(currentStrategy.getType().getQualifiedName() + cache.selectViewStatementType, ret);
	}
	return ret;
}

string Oracle10gObjectStatementsFeature::getSelectFromClassTableStatement(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes, bool useCache)
{
	string stTypePostfix = "@";
	if (useCache) {
		if (attributes.hasFeature("db.select.lock")) {
			stTypePostfix += "L";
		}
		if (attributes.hasFeature("db.select.nowait")) {
			stTypePostfix += "W";
		}
		if (cache.hasCachedStatement(currentStrategy.getType().getQualifiedName() + cache.selectTableStatementType)) {
			return cache.getCachedStatement(currentStrategy.getType().getQualifiedName() + cache.selectTableStatementType + stTypePostfix);
		}
	}

	MetadataHolder tableMeta;
	MetadataHolder selectMeta;
	vector<const MetadataHolder*> cols;
	vector<const MetadataHolder*> tables;
	ostringstream ostr;

	ostr << "SELECT ";

	tableMeta["db.table"].setStringValue(currentStrategy.getType().getMeta("db.table").getStringValue());
	tables.push_back(&tableMeta);

	string type = currentStrategy.getType().getMeta("db.type").getStringValue();
	for (ObjectMappingStrategy::PersistentAttributesIterator it = currentStrategy.getMappedAttributes().begin(); it != currentStrategy.getMappedAttributes().end(); it++) {
		if (it != currentStrategy.getMappedAttributes().begin()) {
			ostr << ", ";
		}
		ostr << "TREAT(VALUE(X) AS " << type << ").";
		ostr << (*it)->getMeta("db.column").getStringValue() << " AS " << (*it)->getMeta("db.column").getStringValue();
	}
	ostr << " FROM " << currentStrategy.getFirstRoot().getType().getMeta("db.table").getStringValue() << " X ";
	ostr << " WHERE " << getWhere(currentStrategy.getDbType().getKeyAttributes(), "db.column");
	// "VALUE(X) IS OF " << type << " AND ";


	if (attributes.hasFeature("db.select.lock")) {
		ostr << " FOR UPDATE";
		if (attributes.hasFeature("db.select.nowait")) {
			ostr << " NOWAIT";
		}
	}

	string ret = ostr.str();
	if (useCache) {
		cache.setCachedStatement(currentStrategy.getType().getQualifiedName() + cache.selectTableStatementType + stTypePostfix, ret);
	}
	return ret;
}

string Oracle10gObjectStatementsFeature::getWhere(const DbType::KeyAttributesSet& keys, string metaColName)
{
	ostringstream out;
	int i = 1;
	for (DbType::KeyAttributesIterator it = keys.begin(); it != keys.end(); it++)
	{
		if (it != keys.begin()) {
			out << " AND ";
		}

		out << (*it)->getMeta(metaColName).getStringValue() << " = :key" << i;
		i++;
	}
	return out.str();
}


/**
 * Creates list of mapped attributes in order that oracle type constructor expects
 *
 * Example:
 * Attributes of types:   child : parent : topMostParent
 * are mapped in order:  topMostParent-parent-child
 */
vector<const MetadataHolder*> Oracle10gObjectStatementsFeature::getColumnsInOrder(ObjectMappingStrategy& strategy)
{
	ObjectMappingStrategy* current = &strategy;
	vector<const MetadataHolder*> cols;

	//cols.push_back(&strategy.getOidColumn());
	while (true) {
		const ObjectMappingStrategy::PersistentAttributesList& attrs = current->getPersistentAttributes();

		vector<const MetadataHolder*> classCols;
		for (ObjectMappingStrategy::PersistentAttributesIterator it = attrs.begin(); it != attrs.end(); it++) {
			classCols.push_back(*it);
		}
		cols.insert(cols.begin(), classCols.begin(), classCols.end());


		if (current->isRoot()) break;
		current = &current->getFirstParentOMStrategy();
	}
	cols.insert(cols.begin(), &strategy.getOidColumn());

	return cols;
}

vector<string> Oracle10gObjectStatementsFeature::getCreateClassTableScript(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes)
{
	vector<string> ret;
	vector<const MetadataHolder*> colMeta;
	vector<const MetadataHolder*> paramsMeta;
	vector<const MetadataHolder*> oidMeta;
	string procParams;
	string oidParam;
	string oidColName = currentStrategy.getOidColumn().getMeta("db.column").getStringValue();

	ostringstream str;

	string typeName = currentStrategy.getType().getMeta("db.type").getStringValue();

	string msg;
	if (!validate(currentStrategy, msg)) {
		IOPC_LOG_THROW(IopcNotSupportedException(msg));
	}


	// Create type

	str << "CREATE OR REPLACE TYPE " << typeName;
	if (currentStrategy.isRoot()) {
		str << " AS OBJECT ( ";
	} else {
		str << " UNDER " << currentStrategy.getType().getFirstParent().getMeta("db.type").getStringValue() << " ( ";
	}

	if (currentStrategy.isRoot()) {
		AdditionalMetadataHolder* amh = new AdditionalMetadataHolder(currentStrategy.getOidColumn());
		string typeSql = typeMapping->getTypeSql(currentStrategy.getOidColumn().getType(), currentStrategy.getOidColumn());
		(*amh)["db.oracle.type.sql"].setStringValue(typeSql);

		if (currentStrategy.getType().isNot<OidObject>()) {
			ostringstream ostr;
			ostr << TypeDesc<OidObject>::getType().getMeta("db.table").getStringValue() <<
				"(" << TypeDesc<OidObject>::getType().getAttribute("oid").getMeta("db.column").getStringValue() << ")";
			(*amh)["db.column.constraint.ref"].setStringValue(ostr.str());
		}

		colMeta.push_back(amh);
	}
	for (ObjectMappingStrategy::PersistentAttributesIterator it = currentStrategy.getPersistentAttributes().begin(); it != currentStrategy.getPersistentAttributes().end(); it++) {
		string typeSql = typeMapping->getTypeSql((*it)->getType(), **it);
		AdditionalMetadataHolder* amh = new AdditionalMetadataHolder(**it);
		(*amh)["db.oracle.type.sql"].setStringValue(typeSql);
		colMeta.push_back(amh);
	}
	str << sqlStatements->getColumns(currentStrategy.getType(), colMeta, true, false, true);

	paramsMeta = getColumnsInOrder(currentStrategy);
	procParams = sqlStatements->getColumns(currentStrategy.getType(), paramsMeta , true, false, true, true, Oracle10gSqlStatementsFeature::TABLE_COLS, "p_");
	oidMeta.push_back(&currentStrategy.getOidColumn());
	oidParam = sqlStatements->getColumns(currentStrategy.getType(), oidMeta , true, false, true, true, Oracle10gSqlStatementsFeature::TABLE_COLS, "p_");

	if (colMeta.size() != 0) {
		str << ", ";
	}
	str << "STATIC PROCEDURE insert_object(" << procParams << ")";
	str << ", STATIC PROCEDURE update_object(" << procParams << ")";
	str << ", STATIC PROCEDURE delete_object(" << oidParam << ")";
	str << " ) NOT FINAL";
	ret.push_back(str.str());

	// Create table

	if (currentStrategy.isRoot()) {
		// Generate table for the root of the object type hierarchy
		ret.push_back(sqlStatements->getCreateTableStatement(currentStrategy.getType(), colMeta));
	}


	AdditionalMetadataHolder indexMeta(currentStrategy.getType(), true);
	indexMeta["db.table"].setStringValue(currentStrategy.getFirstRoot().getType().getMeta("db.table").getStringValue());
	vector<string> indexes = sqlStatements->getIndexes(indexMeta, colMeta);



	for (vector<const MetadataHolder*>::const_iterator it = colMeta.begin(); it != colMeta.end(); it++) {
		delete *it;
	}

	// Crate type body

	// Create the insert/update procedure
	// (Oracle 10g doesn't allow specifying attributes of the type constructor in arbitrary order,
	// thus we create an insert and update procedures (and delete procedure for convenience)

	ostringstream typeBody;

	string constructorParams;
	constructorParams = sqlStatements->getColumns(currentStrategy.getType(), paramsMeta, true, false, false, false, Oracle10gSqlStatementsFeature::TABLE_COLS, "p_");

	typeBody << "CREATE OR REPLACE TYPE BODY " << typeName << " AS ";
	typeBody << "STATIC PROCEDURE insert_object(" << procParams<< ") IS ";
	typeBody << "BEGIN ";
	typeBody << "INSERT INTO " << currentStrategy.getFirstRoot().getType().getMeta("db.table").getStringValue() << " ";
	typeBody << "VALUES(" << typeName << "(" << constructorParams << ")); ";
	typeBody << "END; ";
	typeBody << "STATIC PROCEDURE update_object(" << procParams<< ") IS ";
	typeBody << "BEGIN ";
	typeBody << "UPDATE " << currentStrategy.getFirstRoot().getType().getMeta("db.table").getStringValue() << " X ";
	typeBody << "SET VALUE(X) = NEW " << typeName << "(" << constructorParams << ") ";
	typeBody << "WHERE X." << oidColName << " = p_" << oidColName << "; ";
	typeBody << "END; ";
	typeBody << "STATIC PROCEDURE delete_object(" << oidParam << ") IS ";
	typeBody << "BEGIN ";
	typeBody << "DELETE FROM " << currentStrategy.getFirstRoot().getType().getMeta("db.table").getStringValue() << " ";
	typeBody << "WHERE " << oidColName << " = p_" << oidColName << "; ";
	typeBody << "END; ";
	typeBody << "END;";

	ret.push_back(typeBody.str());

	if (!indexes.empty()) {
		ret.insert(ret.end(), indexes.begin(), indexes.end());
	}

	return ret;
}

vector<string> Oracle10gObjectStatementsFeature::getDropClassTableScript(ObjectMappingStrategy& currentStrategy,  const MetadataHolder& attributes)
{
	vector<string> ret;
	ostringstream dropProcInsert;
	ostringstream dropProcUpdate;
	string typeName = currentStrategy.getType().getMeta("db.type").getStringValue();

/*	dropProcInsert << "DROP PROCEDURE " << typeName << "_insert";
	dropProcUpdate << "DROP PROCEDURE " << typeName << "_update";

	ret.push_back(dropProcInsert.str());
	ret.push_back(dropProcUpdate.str());*/
	ret.push_back(sqlStatements->getDropTableStatement(currentStrategy.getFirstRoot().getType()));
	ostringstream str;
	str << "DROP TYPE " << typeName;
	ret.push_back(str.str());
	return ret;
}

void Oracle10gObjectStatementsFeature::processAttributesForView(ObjectMappingStrategy& currentStrategy, bool& start, ostringstream& ocolsStr, set<string>& processedColumns, const Type& mainType)
{
	// Add oid column not already added
	if (processedColumns.find(currentStrategy.getOidColumn().getMeta("db.viewColumn").getStringValue()) == processedColumns.end()) {
		if (!start) {
			ocolsStr << ", ";
		} else {
			start = false;
		}

		ocolsStr << "TREAT(VALUE(X) AS " << currentStrategy.getType().getMeta("db.type").getStringValue() << ").";
		ocolsStr << currentStrategy.getOidColumn().getMeta("db.column").getStringValue() << " AS " << currentStrategy.getOidColumn().getMeta("db.viewColumn").getStringValue();

		processedColumns.insert(currentStrategy.getOidColumn().getMeta("db.viewColumn").getStringValue());

	}
	for (ObjectMappingStrategy::PersistentAttributesIterator it = currentStrategy.getMappedAttributes().begin(); it != currentStrategy.getMappedAttributes().end(); it++) {
		//if (it->second == DbType::COLSOURCE_INHERITEDPK) continue; // OID will be included from the OidObject
		const Attribute& attr = **it;
		if (!start) {
			ocolsStr << ", ";
		} else {
			start = false;
		}
		ocolsStr << "TREAT(VALUE(X) AS " << currentStrategy.getType().getMeta("db.type").getStringValue() << ").";
		ocolsStr << attr.getMeta("db.column").getStringValue() << " AS " << attr.getMeta("db.viewColumn").getStringValue();

		processedColumns.insert(attr.getMeta("db.viewColumn").getStringValue());
	}
}
