#ifndef LRUCACHE_H_
#define LRUCACHE_H_

// project: POLiTe
// authors: Jan Hadrava 2004

/** \file */
#include <map>

#include "iopccommon/defs.h"

#ifndef WIN32
#include <ext/hash_map>
#endif /*WIN32*/

#include "lruCacheBase.h"
#include "types.h"
#include "iopccommon/condVar.h"


namespace iopc {

template<typename MapType> class LruCache;

typedef LruCache<std::map<PersIdentification, LruCacheItem* > > MapLruCache;

#ifndef WIN32
typedef LruCache<__gnu_cxx::hash_map<PersIdentification, LruCacheItem *> > HashLruCache;
#endif /*WIN32*/

/** Multithreaded cache using LRU replacement policy. Objects are represented by LruCacheItem.
* Items are stored an container conforming to STL "pair associative container" concept. Its type
* is specified by template parameter.
*/
template<typename MapType> class LruCache: public LruCacheBase
{
private:
#ifndef WIN32
	typedef class MapType::iterator Iter;	// MSVC6 doesn't like this although it's perfectly conforming to C++ standard
#else
	typedef MapType::iterator Iter;		// gcc doesn't like this
#endif /*WIN32*/

public:
	inline LruCache(long minValue = MINLONG, long maxCost = MAXLONG, const Strategy& strat = Strategy::defaultStrategy);
	inline LruCache(long minValue, long maxCost, StrategySelector* stratSel);
	inline virtual ~LruCache();

private:
	LruCache(const LruCache<MapType>& src);
	LruCache<MapType> &operator=(const LruCache<MapType>& );

protected:
	virtual void clear(const StrategySelector& strSel);

	virtual CacheLock& makePersistent(const StrategySelector& strSel, const PersIdentification& persId, DatabaseObject& objectPtr);
	virtual CacheLock& getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly = false);

	virtual void dbDelete(const StrategySelector& strSel, const PersIdentification& persId);
	virtual void cacheDelete(const StrategySelector& strSel, const PersIdentification& persId);

	virtual void updateDirty(const StrategySelector& strSel);
	virtual void removeAll(const StrategySelector& strSel);

protected:
	void awaitNonBusy(LruCacheItem &item);
	LruCacheItem& addLock(LruCacheItem& item, bool rdonly);
	void rmLock(LruCacheItem& item);

public:
	virtual inline long getActiveCost();
	virtual inline long getLockedCost();
	virtual long getWorstValue(const TimeStamp& current);
	virtual long removeWorst();
	virtual bool maintainerStep(long& costChange);
	virtual long getSleepTime(const TimeStamp& current, long minValue);
	inline virtual Cache* clone() {
		return new LruCache<MapType>(*this);
	}
protected:
	/** Remove one object scheduled for deletion.
	* \return false there are no objects scheduled for deletion, true otherwise.
	*/
	virtual bool removeDoomed();
	/** Removes one zombie. Zombies are created when an item can't removed
	* from container because of the danger of deadlock.
	* \return false when there are no zombies present, true otherwise.
	*/
	virtual bool removeZombie();
	/** Updates one modified unlocked object.
	* \param costChange Output parameter, the change of total cost caused by method call.
	* \return false when there are no modified unlocked objects present, true otherwise.
	*/
	virtual bool asyncUpdate(long& costChange);
protected:
	MapType cont;		/**< Container holding items of all types - locked/unlocked, representing scheduled actions, modified, history, ... */
	RWLock contLock;	/**< Locks items in container. */
	Mutex listLock;		/**< Common lock needed when modifying lists. */
	CondVar busyCond;	/**< Conditional for signaling state change from ST_BUSY to something else. */
	long activeCnt;		/**< Number of unlocked objects. */
	long lockedCnt;		/**< Number of locked objects. */
	LruCacheItem cleanList; /**< List of objects which doesn't require async update. */
	LruCacheItem dirtyList;	/**< List of objects requiring async update. */
	LruCacheItem doomedList;/**< List of items scheduled for deletion. */
	LruCacheItem zombieList;/**< List of zombies. */
	long activeCost;	/**< Total cost of unlocked items. */
	long lockedCost;	/**< Total cost of locked items. */
};

template<typename MapType> LruCache<MapType>::LruCache(long minValue, long maxCost, StrategySelector* stratSel)
: LruCacheBase(minValue, maxCost, stratSel),
	cont(), contLock(), listLock(), busyCond(),
	activeCnt(0), lockedCnt(0),
	cleanList(), dirtyList(), doomedList(), zombieList(),
	activeCost(0), lockedCost(0)
{
}

template<typename MapType> LruCache<MapType>::LruCache(long minValue, long maxCost, const Strategy& strat)
: LruCacheBase(minValue, maxCost, strat),
	cont(),
	activeCnt(0), lockedCnt(0),
	cleanList(), dirtyList(), doomedList(), zombieList(),
	activeCost(0), lockedCost(0)
{
}

template<typename MapType> LruCache<MapType>::LruCache (const LruCache<MapType>& src)
: LruCacheBase(src.minValue, src.maxCost, src.stratSel, src.defaultStrategy),
	cont(),
	activeCnt(0), lockedCnt(0),
	cleanList(), dirtyList(), doomedList(), zombieList(),
	activeCost(0), lockedCost(0)
{
}

template<typename MapType> LruCache<MapType>::~LruCache()
{
	ExtendedCache::clear();
}

template<typename MapType> void LruCache<MapType>::clear(const StrategySelector& strSel)
{
	IOPC_FUNC_INVOKED;
	contLock.wlock();
	long costChange = 0;
	for (Iter iter = cont.begin(); iter != cont.end(); iter++) {
		LruCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
			contLock.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		long cost;
		switch (item->state) {
		case LruCacheItem::ST_BORN:	/* new item */
			if (item->getObject().isDirty()) {
				try {
					persId.insert(getConnection(), item->getObject());
				}
				IOPC_LOG_CATCH(IopcException& exc)
					contLock.unlock();
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
			}
			item->remove();
			activeCnt--;
			cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
		break;
		case LruCacheItem::ST_ADULT:	/* item with DB copy */
			if (item->getObject().isDirty()) {
				try {
					persId.update(getConnection(), item->getObject());
				}
				IOPC_LOG_CATCH(IopcException& exc)
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
			}
			listLock.lock();
			item->remove();
			activeCnt--;
			cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
			listLock.unlock();
		break;
		case LruCacheItem::ST_DOOMED:	/* item scheduled for deletion */
			try {
				persId.remove(getConnection());
			}
			IOPC_LOG_CATCH(IopcException& exc)
				contLock.unlock();
				IOPC_LOG_THROW(exc);
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
			listLock.lock();
			item->remove();
			listLock.unlock();
		break;
		case LruCacheItem::ST_ZOMBIE:	/* zombie item (busy items can be converted to zombie) */
			listLock.lock();
			item->remove();
			listLock.unlock();
		break;
		case LruCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}

		delete item;
	}
	cont.clear();

	IOPC_ASSERT(activeCnt == 0 && lockedCnt == 0);
	IOPC_ASSERT(activeCost == 0 && lockedCost == 0);
	IOPC_ASSERT(cleanList.prev == &cleanList && cleanList.next == &cleanList);
	IOPC_ASSERT(dirtyList.prev == &dirtyList && dirtyList.next == &dirtyList);
	IOPC_ASSERT(doomedList.prev == &doomedList && doomedList.next == &doomedList);
	IOPC_ASSERT(zombieList.prev == &zombieList && zombieList.next == &zombieList);
	const Strategy& strat = strSel.select();
	contLock.unlock();
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	IOPC_FUNC_ENDED;
}

template<typename MapType> CacheLock& LruCache<MapType>::makePersistent(const StrategySelector &strSel, const PersIdentification& persId, class DatabaseObject& objectPtr)
{
	long costChange;
	contLock.wlock();
	const Strategy& strat = strSel.select(persId);
	LruCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		switch (item->state) {
			case LruCacheItem::ST_BORN:
			case LruCacheItem::ST_ADULT:
				contLock.unlock();
				IOPC_LOG_THROW(IopcDuplicateKeyException());
				break;
			default:
				IOPC_UNREACHABLE;
		}
		LruCacheItem::State oldstate = item->state;
		item->state = LruCacheItem::ST_BUSY;	/* we will keep the item in the cache */
		contLock.unlock();
		switch (oldstate) {
			case LruCacheItem::ST_BORN:
			case LruCacheItem::ST_ADULT:
				IOPC_UNREACHABLE;
				break;
			case LruCacheItem::ST_DOOMED:
				try {
					persId.remove(getConnection());	/* doing maintainer's work */
				}
				IOPC_LOG_CATCH(IopcException& exc)
					busyCond.lock();
					item->state = LruCacheItem::ST_DOOMED;
					busyCond.signal();
					busyCond.unlock();
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
				break;
			case LruCacheItem::ST_ZOMBIE:
				break;
			case LruCacheItem::ST_BUSY:
			default:
				IOPC_UNREACHABLE;
			break;
		}
		listLock.lock();
		item->remove();
		listLock.unlock();
		/* now we have in cache busy item which is prepared to become born/adult */
	} else {
		item = new LruCacheItem(*this, strat, persId, NULL, LruCacheItem::ST_BUSY);
		cont[persId] = item;
		contLock.unlock();
	}
	IOPC_ASSERT(item->state == LruCacheItem::ST_BUSY);
	UpdatingStrategy uStrat = strat.getUpdatingStrategy();
	if (uStrat == USImmed && strat.getLoadLocking()) {
		uStrat = USDeferred;	/* it's not possible to immediately insert object which should be locked in database */
	}
	objectPtr.setDirty(true);
	if (uStrat == USImmed) {
		try {
			persId.insert(getConnection(), objectPtr);
		}
		IOPC_LOG_CATCH(IopcException& exc)
			listLock.lock();
			zombieList.addPrev(*item);
			listLock.unlock();
			costChange = 0;
			busyCond.lock();
			item->state = LruCacheItem::ST_ZOMBIE;
			busyCond.signal();
			busyCond.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT

		objectPtr.setDirty(false);
	}
	item->isLocked = true;
	item->addLock(false);
	//objectPtr.setPersistent(getConnection());
	item->objectPtr = &objectPtr;

	listLock.lock();
	lockedCnt++;
	lockedCost += costChange = strat.cost(item->getObject());
	listLock.unlock();

	busyCond.lock();
	item->state = uStrat == USImmed ? LruCacheItem::ST_ADULT : LruCacheItem::ST_BORN;
	busyCond.signal();
	busyCond.unlock();

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	return *item;
}

template<typename MapType> CacheLock& LruCache<MapType>::getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly)
{
	contLock.wlock();
	const Strategy& strat = strSel.select(persId);
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		LruCacheItem* item = iter->second;
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		switch (item->state) {
		case LruCacheItem::ST_BORN:
		case LruCacheItem::ST_ADULT: {
		/* item is present and active */
			item->addLock(rdonly);
			listLock.lock();
			if (item->isLocked) {
				/* item was already locked,
				** so there's no need to remove it to cleanList
				** and it will be unsafe to reload it */
				listLock.unlock();
				contLock.unlock();
				strat.wakeUpMaintainer(0);
				imposeLimits();
				return *item;
			}
			/* first lock added, item must be added to lockedList and can be reloaded */
			item->remove(); /* removes item from activeList or updateList */
			activeCnt--;
			activeCost -= strat.cost(item->getObject());
			lockedCnt++;
			lockedCost += strat.cost(item->getObject());
			item->isLocked = true;
			listLock.unlock();
			ReadingStrategy rStrat = strat.getReadingStrategy();
			if (item->state == LruCacheItem::ST_BORN || rStrat == RSCache || item->getObject().isDirty()) {
				/* there's no need to reload object */
				contLock.unlock();
				strat.wakeUpMaintainer(0);
				imposeLimits();
				return *item;
			}
			/* object will be reloaded */
			IOPC_ASSERT(item->state == LruCacheItem::ST_ADULT);
			item->state = LruCacheItem::ST_BUSY;
			contLock.unlock(); /* avoid blocking whole cache during reloading */
			try {
				switch (rStrat) {
				case RSTimestamp:
					if (persId.getIDType() == PersIdentification::OID) {
						if (persId.versionMatches(getConnection(), dynamic_cast<OidObject&>(item->getObject()))) {
							break;
						}
					} else {
						LogWriter::logWarning("Can't use reading strategy RSTimestamp with non oid-based object. Using RSTimestamp.");
					}
				/*nobreak*/
				case RSDatabase: {
					DatabaseObject* objectPtr = persId.load(getConnection(), strat.getLoadWait(), strat.getLoadLocking());
					IOPC_ASSERT(objectPtr != NULL);
					item->replaceObject(objectPtr);
				break; }
				case RSCache:
				break;
				default:
					IOPC_UNREACHABLE;
				break;
				}
			}
			IOPC_LOG_CATCH(IopcException&)	/* reload failed, ignore it */
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
			busyCond.lock();
			item->state = LruCacheItem::ST_ADULT;
			busyCond.signal();
			busyCond.unlock();
			strat.wakeUpMaintainer(0);
			imposeLimits();
			return *item;
		break; }
		case LruCacheItem::ST_DOOMED:
		/* item is on the verge of extinction */
			contLock.unlock();
			IOPC_LOG_THROW(IopcException("Item not present"));
		break;
		case LruCacheItem::ST_ZOMBIE:
			listLock.lock();
			item->remove();
			listLock.unlock();
			cont.erase(iter);
			delete item;
		break;
		case LruCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
	}
	/* still keeping contLock */
	IOPC_ASSERT(cont.find(persId) == cont.end());
	LruCacheItem* item = new LruCacheItem(*this, strat, persId, NULL, LruCacheItem::ST_BUSY);
	cont[persId] = item;
	contLock.unlock(); /* item added to cache, contLock can be unlocked */
	DatabaseObject* objectPtr;
	try {
		objectPtr = persId.load(getConnection(), strat.getLoadWait(), strat.getLoadLocking());
	}
	IOPC_LOG_CATCH(IopcException& exc)
		listLock.lock();
		zombieList.addPrev(*item);
		listLock.unlock();
		busyCond.lock();
		item->state = LruCacheItem::ST_ZOMBIE;
		busyCond.signal();
		busyCond.unlock();
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	IOPC_ASSERT(objectPtr != NULL);
	//objectPtr->setPersistent(getConnection());
	item->objectPtr = objectPtr;
	item->addLock(rdonly);
	item->isLocked = true;
	listLock.lock();
	long costChange;
	lockedCnt++;
	lockedCost += costChange = strat.cost(item->getObject());
	listLock.unlock();
	busyCond.lock();
	item->state = LruCacheItem::ST_ADULT;
	busyCond.signal();
	busyCond.unlock();

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	return *item;
}

template<typename MapType> void LruCache<MapType>::dbDelete(const StrategySelector& strSel, const PersIdentification& persId)
{
	long costChange = 0;
	contLock.wlock();
	const Strategy& strat = strSel.select(persId);
	LruCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
			contLock.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		switch (item->state) {
			case LruCacheItem::ST_BORN:	/* remove it immediately */
				listLock.lock();
				item->remove();
				activeCnt--;
				activeCost += costChange = -strat.cost(item->getObject());
				listLock.unlock();
				cont.erase(iter);
				contLock.unlock();
				delete item;
				strat.wakeUpMaintainer(costChange);
				imposeLimits();
				return; /* ignoring the possibility an item already present in the database
					** can have the same key as the removed one */
			break;
			case LruCacheItem::ST_DOOMED:
			/* item is on the verge of extinction */
				contLock.unlock();
				IOPC_LOG_THROW(IopcException("Item not present"));
			break;
			case LruCacheItem::ST_ADULT:
				listLock.lock();
				item->remove();
				activeCnt--;
				activeCost += costChange = -strat.cost(item->getObject());
				listLock.unlock();
				item->release();
				item->state = LruCacheItem::ST_BUSY;
				contLock.unlock();
			break;
			case LruCacheItem::ST_ZOMBIE:
				listLock.lock();
				item->remove();
				item->state = LruCacheItem::ST_BUSY;
				listLock.unlock();
				item->release();	/* release data */
				contLock.unlock();
				costChange = 0;
			break;
			case LruCacheItem::ST_BUSY:
			default:
				IOPC_UNREACHABLE;
			break;
			}
		//IOPC_UNREACHABLE;
	} else {
		item = new LruCacheItem(*this, strat, persId, NULL, LruCacheItem::ST_BUSY);
		cont[persId] = item;
		contLock.unlock();
		costChange = 0;
	}
	IOPC_ASSERT(item->state == LruCacheItem::ST_BUSY);
	if (strat.getUpdatingStrategy() == USImmed) {
		listLock.lock();
		zombieList.addPrev(*item);
		listLock.unlock();
		try {
			persId.remove(getConnection());
		}
		IOPC_LOG_CATCH(IopcException& exc)
			busyCond.lock();
			item->state = LruCacheItem::ST_ZOMBIE;
			busyCond.signal();
			busyCond.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		busyCond.lock();
		item->state = LruCacheItem::ST_ZOMBIE;
		busyCond.signal();
		busyCond.unlock();
	} else { /* USAsync || USImmedDeferred */
		listLock.lock();
		doomedList.addPrev(*item);	/* schedule for deletion */
		listLock.unlock();
		busyCond.lock();
		item->state = LruCacheItem::ST_DOOMED;
		busyCond.signal();
		busyCond.unlock();
	}
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void LruCache<MapType>::cacheDelete(const StrategySelector& strSel, const PersIdentification& persId)
{
	/* ignoring user's advices */
}

template<typename MapType> void LruCache<MapType>::updateDirty(const StrategySelector& strSel)
{
	Iter next;
	contLock.wlock();
	long costChange = 0;
	for (Iter iter = cont.begin(); iter != cont.end(); iter = next) {
		next = iter;
		next++;
		LruCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
			contLock.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		switch (item->state) {
		case LruCacheItem::ST_BORN:	/* new item */
			IOPC_ASSERT(item->getObject().isDirty());
			persId.insert(getConnection(), item->getObject());
			item->getObject().setDirty(false);
			item->state = LruCacheItem::ST_ADULT;
		break;
		case LruCacheItem::ST_ADULT:	/* item with DB copy */
			if (item->getObject().isDirty()) {
				persId.update(getConnection(), item->getObject());
				item->getObject().setDirty(false);
			}
		break;
		case LruCacheItem::ST_DOOMED:	/* item scheduled for deletion */
			persId.remove(getConnection());
			listLock.lock();
			item->remove();
			listLock.unlock();
			cont.erase(iter);
			delete item;
		break;
		case LruCacheItem::ST_ZOMBIE:	/* zombie item (busy items can be converted to zombie) */
			listLock.lock();
			item->remove();
			listLock.unlock();
			cont.erase(iter);
			delete item;
		break;
		case LruCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		if ((item->state == LruCacheItem::ST_BORN || item->state == LruCacheItem::ST_ADULT)
			&& strat.getLoadLocking())
		{	/* item with locking must be removed */
			listLock.lock();
			item->remove();
			activeCnt--;
			long cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
			listLock.unlock();
			delete item;
			cont.erase(iter);
		}
	}
	IOPC_ASSERT(doomedList.next == &doomedList && doomedList.prev == &doomedList);
	IOPC_ASSERT(zombieList.next == &zombieList && zombieList.prev == &zombieList);

	contLock.unlock();
	strSel.select().wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void LruCache<MapType>::removeAll(const StrategySelector& strSel)
{
	long costChange = 0;
	contLock.wlock();
	for (Iter iter = cont.begin(); iter != cont.end(); iter++) {
		LruCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
			contLock.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		listLock.lock();
		item->remove();
		switch (item->state) {
		case LruCacheItem::ST_BORN:
		case LruCacheItem::ST_ADULT:
			activeCnt--;
			activeCost += costChange = -strat.cost(item->getObject());
		break;
		case LruCacheItem::ST_DOOMED:
		case LruCacheItem::ST_ZOMBIE:
		break;
		case LruCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		delete item;
	}
	listLock.unlock();
	cont.clear();
	IOPC_ASSERT(activeCnt == 0 && lockedCnt == 0);
	IOPC_ASSERT(activeCost == 0 && lockedCost == 0);
	IOPC_ASSERT(cleanList.prev == &cleanList && cleanList.next == & cleanList);
	IOPC_ASSERT(dirtyList.prev == &dirtyList && dirtyList.next == & dirtyList);
	IOPC_ASSERT(doomedList.prev == &doomedList && doomedList.next == & doomedList);
	IOPC_ASSERT(zombieList.prev == &zombieList && zombieList.next == & zombieList);

	contLock.unlock();
	strSel.select().wakeUpMaintainer(costChange);
	imposeLimits();
}


template<typename MapType> void LruCache<MapType>::awaitNonBusy(LruCacheItem& item)
{
	busyCond.lock();
	while (item.state == LruCacheItem::ST_BUSY) {
		busyCond.wait();
	}
	busyCond.unlock();
}

template<typename MapType> LruCacheItem &LruCache<MapType>::addLock(LruCacheItem& item, bool rdonly)
{
	if (rdonly) {
		item.lock.rlock();
	} else {
		item.lock.wlock();
	}
	return item;
}

template<typename MapType> void LruCache<MapType>::rmLock(LruCacheItem& item)
{
	/* it's important not to use contLock here so that we avoid deadlock */
	try {
		item.lock.unlock();
	}
	IOPC_LOG_CATCH(IopcException& exc)
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	listLock.lock();
	if (!item.isLocked) {
		listLock.unlock();
		return;
	}
	if (!item.lock.tryXLock()) {
		listLock.unlock();
		return;
	}
	item.remove();	/* remove from list of locked items */
	long cost = item.strat.cost(item.getObject());
	long costChange = 0;
	lockedCnt--;
	lockedCost -= cost;
	item.isLocked = false;
	activeCnt++;
	activeCost += cost;
	UpdatingStrategy uStrat = item.strat.getUpdatingStrategy();
	if (uStrat == USImmed && item.state == LruCacheItem::ST_BORN) {
		IOPC_ASSERT(item.strat.getLoadLocking());
		/* this combination will require inserting item into db and removing local copy but we can't do it
		because it requires state change which can't be made without contLock */
		uStrat = USAsync;
	}
	switch (uStrat) {
	case USDeferred:
		cleanList.addNext(item);
	break;
	case USImmed:
		if (item.state == LruCacheItem::ST_BORN) {
			IOPC_UNREACHABLE;
		} else {
			IOPC_ASSERT(item.state == LruCacheItem::ST_ADULT);
			if (item.getObject().isDirty()) {
				listLock.unlock(); 	/* this is safe because we have XLock for the item */
				item.persId.update(getConnection(), item.getObject());
				item.getObject().setDirty(false);
				listLock.lock();
			}
			cleanList.addNext(item);
		}
	break;
	case USAsync:
		if (item.getObject().isDirty()) {
			dirtyList.addNext(item);
		} else {
			cleanList.addNext(item);
		}
	break;
	}
	item.timeStamp = TimeStamp::now();
	Strategy strat = item.strat;
	item.lock.unlock();
	listLock.unlock();
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> inline long LruCache<MapType>::getActiveCost()
{
	return activeCost;
}

template<typename MapType> inline long LruCache<MapType>::getLockedCost()
{
	return lockedCost;
}

template<typename MapType> long LruCache<MapType>::getWorstValue(const TimeStamp& time)
{
	long retval = MAXLONG;
	listLock.lock();
	if (cleanList.prev != &cleanList) {
		retval = cleanList.prev->timeStamp - time;	/* difference in milliseconds */
	}
	if (dirtyList.prev != &dirtyList) {
		long value = dirtyList.prev->timeStamp - time;
		if (value < retval) {
			retval = value;
		}
	}
	listLock.unlock();
	return retval;
}

template<typename MapType> long LruCache<MapType>::removeWorst()
{
	//LruCacheItem* list = NULL;
	contLock.wlock();
	listLock.lock();
	long value = 0;
	TimeStamp now = TimeStamp::now();
	LruCacheItem* item = NULL;
	if (cleanList.prev != &cleanList) {
		value = cleanList.prev->timeStamp - now;
		item = cleanList.prev;
	}
	if (dirtyList.prev != &dirtyList) {
		if (item == NULL || value > dirtyList.prev->timeStamp - now) {
			item = dirtyList.prev;
		}
	}
	if (item == NULL) {
		listLock.unlock();
		contLock.unlock();
		return 0;
	}
	awaitNonBusy(*item);
	if (item->state == LruCacheItem::ST_BUSY) {
		IOPC_LOG_THROW(IopcIncorrectUsageException("Incorrect behavior of maintainer"));
	}
	bool isAdult;
	if (item->state == LruCacheItem::ST_ADULT) {
		isAdult = true;
	} else {
		IOPC_ASSERT(item->state == LruCacheItem::ST_BORN);
		isAdult = false;
	}
	item->state = LruCacheItem::ST_BUSY;
	item->remove();
	long cost;
	activeCnt--;
	activeCost += cost = -item->strat.cost(item->getObject());
	zombieList.addPrev(*item);
	listLock.unlock();
	contLock.unlock();
	if (item->getObject().isDirty()) {
		if (isAdult) {
			item->persId.update(getConnection(), item->getObject());
		} else {
			item->persId.insert(getConnection(), item->getObject());
		}
		item->getObject().setDirty(false);
	}
	item->release();
	busyCond.lock();
	item->state = LruCacheItem::ST_ZOMBIE;
	busyCond.signal();
	busyCond.unlock();
	return cost;
}

template<typename MapType> bool LruCache<MapType>::removeDoomed()
{
	contLock.wlock();
	listLock.lock();
	if (doomedList.prev == &doomedList) {
		listLock.unlock();
		contLock.unlock();
		return false;
	}
	LruCacheItem& item = *doomedList.prev;
	awaitNonBusy(item);
	IOPC_ASSERT(item.state == LruCacheItem::ST_DOOMED);
	item.state = LruCacheItem::ST_BUSY;
	item.remove();
	zombieList.addPrev(item);
	listLock.unlock();
	contLock.unlock();
	item.persId.remove(getConnection());
	busyCond.lock();
	item.state = LruCacheItem::ST_ZOMBIE;
	busyCond.signal();
	busyCond.unlock();
	return true;
}

template<typename MapType> bool LruCache<MapType>::removeZombie()
{
	contLock.wlock();
	listLock.lock();
	if (zombieList.next == &zombieList) {
		listLock.unlock();
		contLock.unlock();
		return false;
	}
	LruCacheItem *item = zombieList.next;
	awaitNonBusy(*item);
	IOPC_ASSERT(item->state == LruCacheItem::ST_ZOMBIE);
	Iter iter = cont.find(item->persId);
	IOPC_ASSERT(iter != cont.end());
	cont.erase(iter);
	item->remove();
	listLock.unlock();
	contLock.unlock();
	delete item;
	return true;
}

template<typename MapType> bool LruCache<MapType>::asyncUpdate(long& costChange)
{
	costChange = 0;
	contLock.wlock();
	listLock.lock();
	if (dirtyList.prev == &dirtyList) {
		listLock.unlock();
		contLock.unlock();
		return false;
	}
	LruCacheItem &item = *dirtyList.prev;
	LruCacheItem::State oldState = item.state;
	IOPC_ASSERT(oldState == LruCacheItem::ST_BORN || oldState == LruCacheItem::ST_ADULT);
	item.state = LruCacheItem::ST_BUSY;
	item.remove();
	if (oldState == LruCacheItem::ST_BORN && item.strat.getLoadLocking()) {
		/* insert into database and remove from cache */
		zombieList.addPrev(item);
		activeCnt--;
		costChange = -item.strat.cost(item.getObject());
		activeCost += costChange;
		listLock.unlock();
		contLock.unlock();
		item.persId.insert(getConnection(), item.getObject());
		item.release();
		busyCond.lock();
		item.state = LruCacheItem::ST_ZOMBIE;
		busyCond.signal();
		busyCond.unlock();
		return true;
	}
	LruCacheItem *listItem;
	for (listItem = cleanList.next; listItem != &cleanList; listItem = listItem->next) {
		if (listItem->timeStamp <= item.timeStamp) {
			break;
		}
	}
	listItem->addPrev(item);
	IOPC_ASSERT(item.prev == &cleanList || item.prev->timeStamp >= item.timeStamp);
	IOPC_ASSERT(item.next == &cleanList || item.next->timeStamp <= item.timeStamp);
	listLock.unlock();
	contLock.unlock();
	if (oldState == LruCacheItem::ST_ADULT) {
		item.persId.update(getConnection(), item.getObject());
	} else {
		item.persId.insert(getConnection(), item.getObject());
	}
	item.getObject().setDirty(false);
	busyCond.lock();
	item.state = LruCacheItem::ST_ADULT;
	busyCond.signal();
	busyCond.unlock();
	return true;
}

template<typename MapType> bool LruCache<MapType>::maintainerStep(long& costChange)
{
	costChange = 0;
	bool retval = asyncUpdate(costChange);
	if (removeDoomed()) {
		retval = true;
	}
	if (removeZombie()) {
		retval = true;
	}
	return retval;
}

template<typename MapType> long LruCache<MapType>::getSleepTime(const TimeStamp& current, long minValue)
{
	listLock.lock();
	LruCacheItem *item = NULL;
	long diff = 0;

	if (cleanList.prev != &cleanList) {
		diff = cleanList.prev->timeStamp - current;
		item = cleanList.prev;
	}
	if (dirtyList.prev != &dirtyList) {
		long d = dirtyList.prev->timeStamp - current;
		if (item == NULL || diff > d) {
			diff = d;
			item = dirtyList.prev;
		}
	}
	if (item == NULL) {
		listLock.unlock();
///		return 1800*1000000;	/* something like infinity */
		return 5*1000000;
	}
	listLock.unlock();
	if (diff < minValue) {
		return 0;		/* no time to sleep */
	}
	return diff - minValue;
}

}/*namespace iopc*/

#endif /*LRUCACHE_H_*/
