#ifndef CONSTCACHEPTR_H_
#define CONSTCACHEPTR_H_
/**
 * @author Jan Hadrava 2004
 */

#include "cache/cacheLock.h"
#include "persIdentification.h"
#include "db/cachedConnection.h"
#include "iopccommon/countedPtr.h"

namespace iopc {

/** 
 * Template of const cache pointer classes.
 * 
 * Valid const cache pointer references object read-locked in cache (for persistent objects).
 */
template<typename T> class ConstCachePtr{
	
public:
	/** Ctor creating invalid cache pointer */
	ConstCachePtr();
	
	/**
	 * Ctor for creating cache pointer to a persistent object
	 */
	ConstCachePtr(CachedConnection& conn, const PersIdentification& persId);

	/**
	 * Ctor for creating cache pointer to a transient object owned by the DbPtr
	 */
	ConstCachePtr(CountedPtr<T> objectPtr);
	
	/** Ctor creating cache pointer to persistent object */
	ConstCachePtr(CacheLock& cLock);
	
	/** Copy ctor */
	ConstCachePtr(const ConstCachePtr<T>& cptr);
	
	~ConstCachePtr();
	
	/** Copy operator */
	ConstCachePtr<T> &operator=(const ConstCachePtr<T>& src);
	
	/** Invalidates cache pointer */
	void release();
	
	/** Inspects cache pointer validity */
	inline bool isValid() const;
	
	/** Inspects cache pointer validity */
	inline operator bool() const;

	/**
	 * Dereferences cache pointer
	 * 
	 * @return Memory pointer to referenced object (NULL for invalid cache pointer).
	 */
	inline const CountedPtr<T> operator*() const;
	
	/**
	 * Dereferences cache pointer
	 * 
	 * @return Memory pointer to referenced object (NULL for invalid cache pointer).
	 */
	inline const CountedPtr<T> operator->() const;
	
	/**
	 * Dereferences cache pointer
	 *
	 * @return Memory pointer to referenced object (NULL for invalid cache pointer).
	 */
	inline const CountedPtr<T> getObjectPtr() const;
	
protected:
	/** Memory pointer, for both transient and persistent instances. */
	CountedPtr<T> objectPtr;
	
	/** Cache item manipulator (NULL for transient instances). */
	CacheLock* cLock;	
};

template<typename T> ConstCachePtr<T>::ConstCachePtr()
{
	cLock = NULL;
	objectPtr = NULL;
}

template<typename T> ConstCachePtr<T>::ConstCachePtr(CachedConnection& conn, const PersIdentification& persId)
{
	cLock = &conn.getCacheLock(persId, true);
	objectPtr = CountedPtr<T>(dynamic_cast<T*>(&cLock->getObject()), false);
}

template<typename T> ConstCachePtr<T>::ConstCachePtr(CountedPtr<T> objectPtr)
{
	cLock = NULL;
	this->objectPtr = objectPtr;
}

template<typename T> ConstCachePtr<T>::ConstCachePtr(const ConstCachePtr<T>& src)
{
	objectPtr = src.objectPtr;
	if (src.cLock == NULL) {	/* transient instance */
		cLock = NULL;
	} else {			/* persistent instance */
		cLock = &src.cLock->addLock(true);
		IOPC_ASSERT(objectPtr == (T*) &cLock->getObject());
	}
}

template<typename T> ConstCachePtr<T>::ConstCachePtr(CacheLock& cLock)
{
	this->cLock = &cLock;
	objectPtr = CountedPtr<T>(dynamic_cast<T*>(&cLock.getObject()), false);
}

template<typename T> ConstCachePtr<T>::~ConstCachePtr()
{
	IOPC_FUNC_INVOKED;
	if (cLock != NULL) {
		cLock->rmLock();
	}
}

template<typename T> ConstCachePtr<T>& ConstCachePtr<T>::operator=(const ConstCachePtr<T>& src)
{
	if (this == &src)
		return *this;
	objectPtr = src.objectPtr;
	if (cLock != NULL) {
		cLock->rmLock();
	}
	if (src.cLock == NULL) {	/* transient instance */
		cLock = NULL;
	} else {			/* persistent instance */
		cLock = &src.cLock->addLock(true);
		IOPC_ASSERT(objectPtr == (T *)&cLock->getObject());
	}
	return *this;
}

template<typename T> void ConstCachePtr<T>::release()
{
	if (cLock != NULL) {
		cLock->rmLock();
		cLock = NULL;
	}
	objectPtr = NULL;
}

template<typename T> inline bool ConstCachePtr<T>::isValid() const
{
	return !objectPtr.isNull();
}

template<typename T> inline ConstCachePtr<T>::operator bool() const
{
	return isValid();
}

template<typename T> const CountedPtr<T> ConstCachePtr<T>::operator*() const
{
	return getObjectPtr();
}

template<typename T> const CountedPtr<T> ConstCachePtr<T>::operator->() const
{
	return getObjectPtr();
}

template<typename T> const CountedPtr<T> ConstCachePtr<T>::getObjectPtr() const
{
	return objectPtr;
}

} /*namespace*/


#endif /*CONSTCACHEPTR_H_*/
