#include "query.h"
#include "keyValues.h"
#include "types/dbTypeManager.h"
#include "resultBase.h"
#include "iopcdb/driver.h"
#include "db/cachedConnection.h"
#include "iopcdb/cursor.h"

using namespace std;
using namespace iopc;

ResultBase::ResultBase(CachedConnection* conn, const Query& query, bool preload, const Type& type)
 :
		conn(conn),
		query(query.clone()),
		isTransient(type.hasFeature("db.transient")),
		preload(preload && !isTransient),
		type(type),
		dbType(DbTypeManager::getDbType(type))
{

}

ResultBase::~ResultBase()
{
	delete query;
}

void ResultBase::open()
{
	IOPC_FUNC_INVOKED;
	conn->updateDirty();
	//Cursor* cur = conn->sql(query);

	string sqlQuery = query->getSql(conn->getDriver(), type);

	cur = conn->sql(sqlQuery);

	// Only non-transient types which have an db identity can be identified by key
	if (!isTransient) {
		// map key attributes
		int i = 1;
		for (DbType::KeyAttributesIterator it = dbType.getKeyAttributes().begin(); it != dbType.getKeyAttributes().end(); it++)
		{
			const Attribute& attr = **it;
			//TODO: this allows only enhanced-type keys
			EnhancedTypeBase* attrInst = (EnhancedTypeBase*) (attr.getType().newInstance());
			fetchedKey.addNoCopy(attr.getName(), attrInst);
			cur->addEnhancedOutParam(i, *attrInst, attr);
			i++;
		}
	}

	cur->execute();

	if (!preload) {
		if (isTransient) {
			loadCurrentObject(*cur);
		} else {
			fetchNext();
		}
		IOPC_FUNC_ENDED;
		return;
	}

	// preload == true

	while (cur->fetchNext())
	{
		keys.push_back(fetchedKey); // will create copy of fetched data
	}

	cur->close();
	conn->returnCursor(cur);

	IOPC_FUNC_ENDED;
}

bool ResultBase::fetchNext()
{
	if (isTransient) {
		return loadCurrentObject(*cur);
	}

	return (cur->fetchNext());
}

bool ResultBase::fetchPrev()
{
	if (isTransient) {
		IOPC_LOG_THROW(IopcNotSupportedException());
	}
	return (cur->fetchPrev());
}

KeyValues ResultBase::getCurrentKey()
{
	if (isTransient) {
		IOPC_LOG_THROW(IopcNotSupportedException());
	}
	// will create copy of the fetched key
	return fetchedKey;
}

KeyValues ResultBase::getKeyAt(int i)
{
	if (isTransient) {
		IOPC_LOG_THROW(IopcNotSupportedException());
	}
	return keys[i];
}


void ResultBase::close()
{
	if (!preload && cur->isOpened()) {
		cur->close();
		conn->returnCursor(cur);
	}
	keys.clear();
}
