#ifndef ENHANCEDSTRING_H_
#define ENHANCEDSTRING_H_
#include <string>
#include "enhancedTypeBase.h"
#include "iopccommon/defs.h"
#include "typeDesc.h"
#include <sstream>
#include <iostream>

namespace iopc {

template<typename T> class EStr;


template<typename T> EStr<T> operator+ (const T* str, const EStr<T>& str2)
{
	return EStr<T>(str + str2.value);
}

template<typename T> EStr<T> operator+ (const std::basic_string<T>& str, const EStr<T>& str2)
{
	return EStr<T>(str + str2.value);
}


template<typename T>
std::ostream& operator<<(std::ostream& o, const EStr<T>& str)
{
	o << str.value;
	return o;
}

template<typename T> class DLL_EXTERNAL EStr : public EnhancedTypeBase {
private:
	std::basic_string<T> value;
	T* copy;
	unsigned int bufferLen;

public:
	//ENumber();
	EStr(const std::basic_string<T>& v);
	EStr(const EStr<T>& src);
	EStr(const T* str);
	EStr();

	virtual ~EStr();

	// Operators
	operator std::basic_string<T>() const;
	//operator T&();  // Do not allow obtaining reference to wrapped value

	EStr<T>& operator= (const T* str);
	EStr<T>& operator= (const EStr<T>& src);
	EStr<T>& operator= (const std::basic_string<T>& val);

	EStr<T> operator+ (const EStr<T>& src);
	EStr<T> operator+ (const T* str);

	int operator!= (const EStr<T>& src);
//	int operator!= (const std::basic_string<T>& val);
	int operator== (const EStr<T>& src);
//	int operator== (const std::basic_string<T>& val);

	template<typename U>
	friend EStr<U> operator+(const U*, const EStr<U>& str2);

	template<typename U>
	friend EStr<U> operator+(const std::basic_string<U>&, const EStr<U>& str2);

  	virtual const Type& getType() const;
  	virtual const TypeDesc<EStr<T> >& getTypeDesc() const;

  	/**
	 * Creates a copy of this object's value and returns a pointer to it.
	 */
	virtual void* externalizeCopy(unsigned int maxlength);

	/**
	 * Creates an externalized empty buffer for obtaining the value.
	 */
	virtual void* externalizeEmptyBuffer(unsigned int length);

	/**
	 * Updates this object's value from the externalized buffer
	 */
	virtual void updateFromExternalized();

  	/**
  	 * Returns externalized buffer's size in bytes
  	 *
  	 * If buffer was not externalised, returns 0
  	 */
  	virtual unsigned int getBufferSize();

  	virtual const Type& getInnerType() const;

  	virtual std::string toString() const;

  	virtual Object* clone() const;

  	virtual int compare(const Object& o) const;

  	template<typename U>
  	friend std::ostream& operator<<(std::ostream& out, const EStr<U>& str);

  	//template<typename U>
  	friend class iopc::TypeDesc<EStr<T> >;
};

template<typename T> int EStr<T>::compare(const Object& o) const
{
	const Type& type1 = TypeManager::getTypeByTypeInfo(&typeid(*this));
	const Type& type2 = TypeManager::getTypeByTypeInfo(&typeid(o));

	if (type1 != type2) {
		IOPC_LOG_THROW(IopcIncorrectUsageException(std::string("Can't compare EString with type ") + type2.getQualifiedName()));
	}
	std::basic_string<T> cmpVal = (dynamic_cast<const EStr<T>& >(o)).value;

	if (value < cmpVal) {
		return -1;
	} else if (value > cmpVal) {
		return 1;
	} else {
		return 0;
	}
}


template<typename T> Object* EStr<T>::clone() const
{
	return new EStr<T>(*this);
}

template<typename T> const Type& EStr<T>::getType() const
{
	return TypeDesc<EStr<T> >::getType();
}

template<typename T> const TypeDesc<EStr<T> >& EStr<T>::getTypeDesc() const
{
	return TypeDesc<EStr<T> >::getType();
}

template<typename T> EStr<T>::EStr(const std::basic_string<T>& v)
{
	value = v;
	copy = NULL;
	bufferLen = 0;
	resetNull();
}

template<typename T> EStr<T>::EStr(const EStr<T>& src)
{
	value = src.value;
	setNull(src.isNull());
	copy = NULL;
	bufferLen = 0;
}

template<typename T> EStr<T>::EStr(const T* str)
{
	value = str;
	copy = NULL;
	bufferLen = 0;
	resetNull();
}

template<typename T> EStr<T>::EStr()
{
	copy = NULL;
	bufferLen = 0;
}

template<typename T> EStr<T>::~EStr()
{
	if (copy) delete[] copy;
}

template<typename T> EStr<T>::operator std::basic_string<T>() const
{
	return value;
}

template<typename T> EStr<T>& EStr<T>::operator=(const std::basic_string<T>& val)
{
	value = val;
	setDirty();
	resetNull();

	return *this;
}

template<typename T> EStr<T>& EStr<T>::operator=(const EStr<T>& src)
{
	if (isNull() && src.isNull()) return *this;
	setNull(src.isNull());
	setDirty();
	if (isNull()) {
		return *this;
	}
	value = src.value;

	return *this;
}

template<typename T> EStr<T>& EStr<T>::operator=(const T* str)
{
	value = str;
	setDirty();
	resetNull();

	return *this;
}

template<typename T> EStr<T> EStr<T>::operator+ (const T* str)
{
	return EStr<T>(value + str);
}

template<typename T> EStr<T> EStr<T>::operator+ (const EStr<T>& src)
{
	return EStr<T>(value + src.value);
}

template<typename T> int EStr<T>::operator!= (const EStr<T>& src)
{
	return value != src.value;
}

template<typename T> int EStr<T>::operator== (const EStr<T>& src)
{
	return value == src.value;
}
/*
template<typename T> int EStr<T>::operator!= (const std::basic_string<T>& val)
{
	return value != val;
}

template<typename T> int EStr<T>::operator== (const std::basic_string<T>& val)
{
	return value == val;
}
*/

template<typename T> void* EStr<T>::externalizeCopy(unsigned int maxlength)
{
	if (copy) delete[] copy;
	copy = new T[maxlength + 1];
	int chars = value.copy(copy, maxlength);
	copy[chars] = '\0';
	bufferLen = maxlength;
	return copy;
}

template<typename T> void EStr<T>::updateFromExternalized()
{
	value = copy;
	setDirty();
	resetNull();
}

template<typename T> void* EStr<T>::externalizeEmptyBuffer(unsigned int length)
{
	if (copy) delete[] copy;
	copy = new T[length + 1];
	bufferLen = length;
	return copy;
}

template<typename T> unsigned int EStr<T>::getBufferSize()
{
	if (!copy) return 0;
	return (bufferLen + 1)*sizeof(T);
}

template<typename T> const Type& EStr<T>::getInnerType() const
{
	return TypeDesc<std::basic_string<T> >::getType();
}

typedef EStr<char> EString;
typedef EStr<wchar_t> EWString;

///////////// Type descriptors ///////////////


template <> class DLL_EXTERNAL TypeDesc<iopc::EString> : public Type {
public:
	inline virtual DataTypeEnum getDataType() const { return IOPC_TYPE_ESTRING; }
	inline virtual const std::string getName() const { return "EString"; }
	inline virtual const std::string getNamespace() const { return "iopc"; }
	inline virtual void* newInstance() const { return new iopc::EString(); }
	inline virtual Object* newObjectInstance() const { return new iopc::EString(); }
	inline virtual int getSizeOf(const void* val) const { return (((const EString*) val)->value.length() + 1) * sizeof(char); }
	static inline const TypeDesc<iopc::EString>& getType() { return instance; }
	inline virtual DataTypeClassEnum getDataTypeClass() const { return IOPC_TYPECLASS_ENHANCED; }
	inline virtual void fillAttributeValues(const Object& o, std::map<std::string, AttributeValue>& out) const {}
	inline virtual bool isBuiltIn() const { return true; }
	inline virtual void deleteInstance(void* instance) const { delete (iopc::EString*) instance; }
private:
	TypeDesc() { TypeManager::registerType(this, "iopc::EString", &typeid(iopc::EString)); }
	static TypeDesc<iopc::EString> instance;
};

template <> class DLL_EXTERNAL TypeDesc<iopc::EWString> : public Type {
public:
	inline virtual DataTypeEnum getDataType() const { return IOPC_TYPE_EWSTRING; }
	inline virtual const std::string getName() const { return "EWString"; }
	inline virtual const std::string getNamespace() const { return "iopc"; }
	inline virtual void* newInstance() const { return new iopc::EWString(); }
	inline virtual Object* newObjectInstance() const { return new iopc::EWString(); }
	inline virtual int getSizeOf(const void* val) const { return (((const EWString*) val)->value.length() + 1) * sizeof(wchar_t); }
	static inline const TypeDesc<iopc::EWString>& getType() { return instance; }
	inline virtual DataTypeClassEnum getDataTypeClass() const { return IOPC_TYPECLASS_ENHANCED; }
	inline virtual void fillAttributeValues(const Object& o, std::map<std::string, AttributeValue>& out) const {}
	inline virtual bool isBuiltIn() const { return true; }
	inline virtual void deleteInstance(void* instance) const { delete (iopc::EWString*) instance; }
private:
	TypeDesc() { TypeManager::registerType(this, "iopc::EWString", &typeid(iopc::EWString)); }
	static TypeDesc<iopc::EWString> instance;
};

//SIMPLE_TYPE_DEFINITION("EString", EString, "iopc", "iopc::EString", IOPC_TYPE_ESTRING, IOPC_TYPECLASS_ENHANCED);
//SIMPLE_TYPE_DEFINITION("EWString", EWString, "iopc", "iopc::EWString", IOPC_TYPE_EWSTRING, IOPC_TYPECLASS_ENHANCED);




}

#endif /*ENHANCEDSTRING_H_*/
