/**
 * @file metadataInitializer.h
 *
 * @ingroup iopcmeta
 *
 * @author Petr Cermak
 */

#ifndef METADATAINITIALIZER_H_
#define METADATAINITIALIZER_H_

namespace iopc {

/**
 * @brief Allows users to initialize class metadata at the library startup.
 *
 * Class metadata can be initialized in two standard ways:
 * either by implementing the static void iopcInit(iopc::Type& t)
 * or by passing an instance of a MetadataInitializer implementation
 * to @a IopcMeta::init(MetadataInitializer* metadataInitializer)
 * Only these two ways guarantee that the class metadata will be set properly
 * and that they will have desired effect on the program execution.
 *
 * Example:
 * @code
 * class OurMetadataInitializer : public MetadataInitializer {
 * public:
 * virtual void initializeMetadata() {
 *   globalSettings["db.oracle.oidSequence.name"].setStringValue("OIDS");
 *   TypeDesc<MyType>::getType()["db.table"].setStringValue("MYTABLE");
 *   ...
 * };
 *
 * ...
 *
 * int main(int argc, char *argv[])
 * {
 *   Iopc::init(new OurMetadataInitializer(), LogWriter::ERROR ... );
 *   ...
 * }
 * @endcode
 *
 * @ingroup iopcmeta
 */
class MetadataInitializer {
	public:

		/**
		 * @brief Implementations should set the class metadata or global settings here
		 *
		 * Typical implementation would load metadata from a configuration file
		 * and use TypeManager and related classes to set them on matching classes and attributes.
		 */
		virtual void initializeMetadata()=0;

		virtual ~MetadataInitializer() {};
};


}

#endif /* METADATAINITIALIZER_H_ */
