#ifndef TYPE_H_
#define TYPE_H_

#include "iopccommon/defsPlatform.h"
#include "object.h"
#include "attribute.h"
#include "attributeValue.h"
#include "iopccommon/metadataHolder.h"
//#include "meta/typeDesc.h"
#include <string>
#include <map>
#include <vector>
#include <set>

namespace iopc {

class IopcCommon;
class Object;

#define NAMESPACE_DELIMITER "::"

template<typename T> class TypeDesc;

class DLL_EXTERNAL Type : public MetadataHolder {

	friend class TypeManager;

public:
	virtual const std::string getName() const = 0;
	virtual const std::string getNamespace() const = 0;
	inline const std::string getQualifiedName() const {
		return getNamespace() + NAMESPACE_DELIMITER + getName();
	}

	virtual int getSizeOf(const void* object) const =0;
	virtual bool isBuiltIn() const = 0;

	virtual void* newInstance() const = 0;
	virtual Object* newObjectInstance() const =0;
	virtual void deleteInstance(void* instance) const = 0;

	enum DataTypeClassEnum {
		IOPC_TYPECLASS_SIMPLE = 0,
		IOPC_TYPECLASS_STRING,
		IOPC_TYPECLASS_COMPLEX,
		IOPC_TYPECLASS_ENHANCED
	};

	virtual DataTypeClassEnum getDataTypeClass() const = 0;

	enum DataTypeEnum {
	  IOPC_TYPE_INVALID = 0, //!< not a valid type

	  IOPC_TYPE_BOOL,        //!< bool

	  IOPC_TYPE_CHAR,        //!< char (signed or unsigned - depends on platform and compiler settings)
	  IOPC_TYPE_UCHAR,       //!< unsigned char
	  IOPC_TYPE_SCHAR,       //!< signed char
	  IOPC_TYPE_SHORT,       //!< short (signed)
	  IOPC_TYPE_USHORT,      //!< unsigned short
	  IOPC_TYPE_INT,         //!< int (signed)
	  IOPC_TYPE_UINT,        //!< unsigned
	  IOPC_TYPE_LONG,        //!< long (signed)
	  IOPC_TYPE_ULONG,       //!< unsigned long

	  IOPC_TYPE_FLOAT,       //!< float
	  IOPC_TYPE_DOUBLE,      //!< double
	  IOPC_TYPE_LONGDOUBLE,  //!< long double

	  IOPC_TYPE_WCHART,     //!< wchar_t

	  IOPC_TYPE_STRING,      //!< char * (0-terminated)
	  IOPC_TYPE_WSTRING,     //!< wchar_t * (00-terminated)

	  IOPC_TYPE_REF,         //!< reference to some other persistent object

	  IOPC_TYPE_COMPLEX,   //!< some complex data type (array, class ...)

  	  IOPC_TYPE_EINT,
  	  IOPC_TYPE_EUINT,
  	  IOPC_TYPE_ECHAR,
  	  IOPC_TYPE_EUCHAR,
  	  IOPC_TYPE_ESCHAR,
  	  IOPC_TYPE_ESHORT,
  	  IOPC_TYPE_EUSHORT,
  	  IOPC_TYPE_ELONG,
  	  IOPC_TYPE_EULONG,
  	  IOPC_TYPE_EFLOAT,
  	  IOPC_TYPE_EDOUBLE,
  	  IOPC_TYPE_ELDOUBLE,
  	  IOPC_TYPE_EWCHART,
  	  IOPC_TYPE_EBOOL,

  	  IOPC_TYPE_ESTRING,
  	  IOPC_TYPE_EWSTRING
	};

	virtual DataTypeEnum getDataType() const =0;

	typedef std::vector<const Type*> ParentsList;
	typedef std::vector<const Type*>::const_iterator ParentsIterator;
	typedef std::vector<const Type*> ChildrenList;
	typedef std::vector<const Type*>::const_iterator ChildrenIterator;
	typedef std::set<const Type*> AllParentsList;
	typedef std::set<const Type*>::const_iterator AllParentsIterator;
	typedef std::set<const Type*> AllChildrenList;
	typedef std::set<const Type*>::const_iterator AllChildrenIterator;
	typedef std::map<std::string, Attribute> AttributesMap;
	typedef AttributesMap::const_iterator AttributesIterator;
	typedef std::map<std::string, const Attribute*> InheritedAttributesMap;
	typedef InheritedAttributesMap::const_iterator InheritedAttributesIterator;

	virtual void fillAttributeValues(const Object& o, std::map<std::string, AttributeValue>& out) const =0;
	inline const ParentsList& getParents() const { return parents; };
	inline const ChildrenList& getChildren() const { return children; };
	inline const AllChildrenList& getAllChildren() const { return allChildren; };
	inline const AllParentsList& getAllParents() const { return allParents; }
	inline const AttributesMap& getAttributes() const { return attributes; };
	inline const InheritedAttributesMap& getInheritedAttributes() const { return inheritedAttributes; }
	const Attribute& getAttribute(const std::string& name) const;
	bool hasAttribute(const std::string& name) const;
	const Attribute& getAttributeAll(const std::string& name) const;
	bool hasAttributeAll(const std::string& name) const;

	/**
	 * Assumes that this type has only one parent and returns it
	 *
	 * If this type represents the iopc::Object type, returns NULL
	 */
	const Type& getFirstParent() const;

	bool isTypeOf(const Type& type) const;
	template<typename T> bool isTypeOf() const {
		return isTypeOf(TypeDesc<T>::getType());
	}

	template<typename T> bool is() const {
		return *this == TypeDesc<T>::getType();
	}

	template<typename T> bool isNot() const {
		return *this != TypeDesc<T>::getType();
	}


public:
	inline bool operator!=(const Type& src) const { return this != &src; }
	inline bool operator==(const Type& src) const { return this == &src; }
	inline bool operator<(const Type& src) const { return this < &src; }

protected:

	Type();
	virtual ~Type();

	virtual void iopcInit() {};

	AttributesMap attributes;
	InheritedAttributesMap inheritedAttributes;
	std::vector<const std::type_info*> parentsTypeInfo;
	ParentsList parents;
	ChildrenList children;
	AllChildrenList allChildren;
	AllParentsList allParents;

	/**
	 * Cached set of all (even indirect) parents.
	 */

private:
	void generateAllChildren();
	void generateAllParents();
	void generateInheritedAttributes();

	/**
	 * Called via TypeManager from a friend class IopcCommon::init()
	 */
	void init();
	void postInit();


	// TODO odstranit - pro debug
	std::string typeName;
};

}

#endif /*TYPE_H_*/
