#ifndef TYPEMANAGER_H_
#define TYPEMANAGER_H_

#include "type.h"
#include <map>
#include <string>
#include <typeinfo>
#include <vector>
#include "iopccommon/lock.h"

namespace iopc {


/**
 * Predicate for comparing two pointers to type_info
 */
struct typeInfoCmp
{
	bool operator()(const std::type_info *t1, const std::type_info *t2) const {
		return t1->before(*t2) != 0;
	}
};


class DLL_EXTERNAL TypeManager {
public:
	typedef std::map<std::string, const Type*> TypesMap;
	typedef TypesMap::const_iterator TypesIterator;


	static TypeManager& getInstance();

	static inline void registerType(const Type* type, const std::type_info* ti) {
		getInstance().registerTypeInst(type, ti);
	}

	static inline void registerType(const Type* type, const std::string& typeName, const std::type_info* ti) {
		getInstance().registerTypeInst(type, typeName, ti);
	}

	/*static inline std::map<std::string, const Type*>::iterator getTypesIterator() {
		return getInstance().getTypesIteratorInst();
	}*/

	/**
	 * Returns a type object that represents a type
	 * specified by its name.
	 *
	 * The typeName parameter must be a fully qualified name.
	 * Example: "iopc::FakeType", "::TopLevelType"
	 *
	 * @param typeName A fully qualified type name.
	 * @return A type object that represents the type.
	 */
	static inline const Type& getType(const std::string& typeName) {
		return getInstance().getTypeInst(typeName);
	}

	static inline bool hasType(const std::string& typeName) {
		return getInstance().hasTypeInst(typeName);
	}

	/**
	 * Returns a type object that represents a type
	 * specified by its type_info object.
	 *
	 * This object is returned by the typeid operator.
	 *
	 * @param ti A type_info object
	 * @return A type object that represents the type.
	 */
	static inline const Type& getTypeByTypeInfo(const std::type_info* ti) {
		return getInstance().getTypeByTypeInfoInst(ti);
	}

	void registerTypeInst(const Type* type, const std::type_info* ti);
	void registerTypeInst(const Type* type, const std::string& typeName,const  std::type_info* ti);

	/**
	 * Returns a type object that represents a type
	 * specified by its name.
	 *
	 * The typeName parameter must be a fully qualified name.
	 * Example: "iopc::FakeType", "::TopLevelType"
	 *
	 * @param typeName A fully qualified type name.
	 * @return A type object that represents the type.
	 */
	const Type& getTypeInst(const std::string& typeName);

	bool hasTypeInst(const std::string& typeName);

	/**
	 * Returns a type object that represents a type
	 * specified by its type_info object.
	 *
	 * This object is returned by the typeid operator.
	 *
	 * @param ti A type_info object
	 * @return A type object that represents the type.
	 */
	const Type& getTypeByTypeInfoInst(const std::type_info* ti);


	static const TypesMap& getTypesMap() {
		return getInstance().getTypesMapInst();
	}

	/**
	 * Initializes type instances.
	 *
	 * Called from iopcCommon::init()
	 */
	static inline void init() {
		getInstance().initInst();
	}

	//std::map<std::string, const Type*>::iterator getTypesIteratorInst();
	const TypesMap& getTypesMapInst() const;

	static inline void topologicalSort(const Type& root, std::vector<const Type*>& sorted) {
		getInstance().topologicalSortInst(root, sorted);
	}
	void topologicalSortInst(const Type& root, std::vector<const Type*>& sorted);

	/**
	 * Initializes type instances.
	 *
	 * Called from iopcCommon::init()
	 */
	void initInst();


	void parseTypeInst(std::string inStr,  std::string& type,  std::string& typeMeta,  std::string& attr, std::string& attrMeta);
	void parseMetaInst(std::string inStr,  std::string& outStr,  std::string& outMeta);

	static inline void parseType(std::string inStr,  std::string& type,  std::string& typeMeta,  std::string& attr, std::string& attrMeta) {
		getInstance().parseTypeInst(inStr, type, typeMeta, attr, attrMeta);
	}

	static inline void parseMeta(std::string inStr,  std::string& outStr,  std::string& outMeta) {
		getInstance().parseMetaInst(inStr, outStr, outMeta);
	}

protected:
	TypeManager();
	~TypeManager();

private:
	std::map<std::string, const Type*> types;
	std::map<const std::type_info*, const Type*, struct typeInfoCmp> typeInfoToType;

	Lock lock;
};


}

#endif /*TYPEMANAGER_H_*/
